/***************************************************************************
                     AC3_audio_module.c  -  description
                             -------------------
   AC3 audio PES module functions.
                             -------------------
    begin                : Thu Aug 18 2005
    copyright            : (C) 2005 by Dennis Lew
    email                : dennis_L76@yahoo.ca
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
/*#define AC3_AUDIO_MODULE_TRACE*/
#ifdef AC3_AUDIO_MODULE_TRACE
# include <stdio.h>
#endif
#include  <string.h>

#include "PES_PKT.h"
#include "commands.h"
#include "sound/sound.h"
#include "AC3_audio_module.h"
#include "AC3_audio.h"
#include "audio_block.h"
#include "imdft.h"
#include "window.h"

typedef struct {
   AC3_SYNC_FRAME frame;
#ifdef ALIGN_BYTES
   u8 unused1[ALIGN_BYTES - (sizeof(AC3_SYNC_FRAME)&(ALIGN_BYTES-1))];
#endif
   AUDIO_BLOCK    blk;
   u32            sound_init;
   s32            frame_interval;
   SYS_DATA      *sys_data;
} AC3_audio_module_data;

static void init_audio_buf(AC3_SYNC_FRAME *frame, AUDIO_BLOCK *blk, u32 num_ch);
static void output_block(AC3_SYNC_FRAME *frame, AUDIO_BLOCK *blk, u32 num_ch);

/*****************************************************************
   GENERAL MODULE FUNCTIONS
     These functions are used by all SYSTEM, PSI, and PES modules.
 *****************************************************************/
/* Returns the number of bytes of memory needed by the module. */
size_t AC3_AUDIO_MODULE_mem_needed()
{
   return sizeof(AC3_audio_module_data);
}

/* Initializes the memory provided for the module.
      INPUT: void *     - Pointer to the memory provided for the module.
             SYS_DATA * - Pointer to system data. */
void AC3_AUDIO_MODULE_init(void *mem, SYS_DATA *sys_data)
{
   AC3_audio_module_data *data = (AC3_audio_module_data *)mem;

   data->frame_interval = 0;
   data->frame.state    = AC3_AUDIO_next_frame_STATE;
   
   if (sys_data) {
/*      if (!data->sys_data) {
      printf("data          =0x%08X\n"
             "data->frame   =0x%08X\n"
             "data->blk     =0x%08X\n"
             "blk->ch_data  =0x%08X 0x%08X 0x%08X 0x%08X 0x%08X\n"
             "blk->lfe_data =0x%08X\n"
             "blk->delay_buf=0x%08X\n",
             (u32)data,
             (u32)&(data->frame),
             (u32)&(data->blk),
             (u32)&(data->blk.ch_data[0].coeff),(u32)&(data->blk.ch_data[1].coeff),
             (u32)&(data->blk.ch_data[2].coeff),(u32)&(data->blk.ch_data[3].coeff),
             (u32)&(data->blk.ch_data[4].coeff),
             (u32)&(data->blk.lfe_data.coeff),
             (u32)&(data->blk.delay_buf));
      }*/
      data->sys_data      = sys_data;
      sys_data->a_sync    = 0;
      data->frame.a_sync  = &(sys_data->a_sync);
      data->frame.pes_pts = 0;
   }
}

/* Clean up (DO NOT FREE) the memory provided for the module.
      INPUT: void * - Pointer to the memory provided for the module. */
void AC3_AUDIO_MODULE_done(void *mem)
{
   AC3_audio_module_data *data = (AC3_audio_module_data *)mem;
   
   if (data->sound_init) {
      data->sound_init = 0;
      SOUND_done();
   }

   data->frame_interval = 0;
   data->frame.state    = AC3_AUDIO_next_frame_STATE;

   if (data->sys_data) {
      data->sys_data->a_sync = 0;
      data->frame.pes_pts    = 0;
   }
}

/* Gives the module the offset value needed to obtain the program clock value.
      INPUT: void *  - Pointer to the memory provided for the module.
             CLK_VAL - offset value */
void AC3_AUDIO_MODULE_set_offset(void *mem, CLK_VAL offset)
{
   AC3_audio_module_data *data = (AC3_audio_module_data *)mem;
   
   data->sys_data->a_sync = 0;
   data->frame.offset     = offset;
   data->frame.pes_pts    = 0;
}

/* Alerts the module to a discontinuity in the data stream.
      INPUT: void * - Pointer to the memory provided for the module. */
void AC3_AUDIO_MODULE_discontinuity(void *mem)
{
   AC3_audio_module_data *data = (AC3_audio_module_data *)mem;

   data->frame.state = AC3_AUDIO_next_frame_STATE;

   if (data->sys_data) {
      data->sys_data->a_sync = 0;
      data->frame.pes_pts    = 0;
   }
}


/**************************************************************************
   PES MODULE FUNCTIONS
     These functions are used by SYSTEM modules to process the  PES_PKTs
       they produce.
 **************************************************************************/
/* Initialize everything needed to handle the given elementary stream.
     The PES module will handle the elementary stream (if able) until the
     MPEG2_MODULE_done function is called, or this function is called
     again.
      INPUT: void *        - Pointer to the memory provided for the module.
             PMT_SECTION * - Pointer to the PMT data.
             int           - Index, in elem array, of the elementary stream
                               that needs to be handled.
      RETURN: Module is ready to handle
                the stream.               - Non-zero value.
              Module cannot handle the
                stream.                   - 0 */
int AC3_AUDIO_MODULE_stream_init(void *mem, PMT_SECTION *pmt, int idx)
{
   return (pmt->elem[idx]->stream_type == STREAM_TYPE_ATSC_AC3_AUDIO);
}

/* Processes the packet handled by the module.
      INPUT: void * - Pointer to the memory provided for the module.
             void * - Pointer to the PES_PKT that needs handling.
                        Set pes->stream_id to zero to let the PES
                        module know to expect the start of a new
                        PES_PKT. */
void AC3_AUDIO_MODULE_process_pkt(void *mem, void *pes)
{
   AC3_audio_module_data *data = (AC3_audio_module_data *)mem;
   PES_PKT               *pkt  = (PES_PKT *)pes;

#ifndef _SPEED_TEST_
   if (PES_PTS_FLAG(pkt))
      data->frame.pes_pts = (pkt->PTS) ? pkt->PTS : 1;
#endif

#ifdef AC3_AUDIO_MODULE_TRACE
   printf("start AC3_AUDIO_MODULE_process_pkt(%hu)\n", pkt->payload_len16);
#endif
   while (pkt->payload_len16) {
      if (data->frame.state < AC3_AUDIO_next_frame_END_STATE) {
         /* look for a new audio frame */
         AC3_AUDIO_next_frame(pkt->payload,
                              (u32)pkt->payload_len16,
                              &(data->frame));
         if (data->frame.state < AC3_AUDIO_next_frame_END_STATE)
            break;
#ifdef AC3_AUDIO_MODULE_TRACE
         
         printf("<START_FRAME>\n"
                "  state=0x%04hX  bytes=%u\n", 
                data->frame.state, data->frame.frame_len);
#endif
         pkt->payload_len16 = data->frame.payload_len16;
         data->blk.num      = 0;
         if (!data->frame_interval) {
            data->frame_interval = (s32)((AC3_NUM_BINS*AC3_BLKS_PER_FRAME*MPEG2_SYS_CLK_FREQ)/
                                         data->frame.sample_rate);
         }
         data->sys_data->a_pos += (CLK_VAL)data->frame_interval;

         if ( (data->sys_data->a_skip > 0)||
              (data->sys_data->command == CMD_SLOW_FWD) ) {
            if (data->sys_data->a_skip > 0)
               data->sys_data->a_skip -= (CLK_VAL)data->frame_interval;
            data->sys_data->a_sync = 0;
            data->frame.state      = AC3_AUDIO_SKIP_AUX_BYTES; /* don't play audio */
            pkt->payload           = data->frame.payload;
            SOUND_stop(0);
         }
         else if (!data->sound_init) {
            data->sound_init = SOUND_init((u32)data->frame.num_ch, 
                                          data->frame.sample_rate);
            if (data->sound_init) {
               data->blk.audio_buf[0] = SOUND_buf(AC3_NUM_BINS);
               if (data->blk.audio_buf[0]) {
                  init_audio_buf(&(data->frame), &(data->blk), data->sound_init);
                  goto AC3_AUDIO_MODULE_process_pkt_read_BSI;
               }
            }
            data->sound_init  = 0;
            data->frame.state = AC3_AUDIO_next_frame_STATE;
         }
         else
            goto AC3_AUDIO_MODULE_process_pkt_read_BSI;
      }
      else {   
#ifdef AC3_AUDIO_MODULE_TRACE
         printf("  state=0x%04hX  bytes=%u\n", 
                data->frame.state, data->frame.frame_len);
#endif
         /* deal with the payload */
         data->frame.payload_len16 = pkt->payload_len16;
         data->frame.payload       = pkt->payload;

         if (data->frame.state >= AUDIO_BLOCK_read_config_STATE)
            goto AC3_AUDIO_MODULE_process_pkt_LOOP;
         else {
         /*if (data->frame.state < AC3_AUDIO_read_BSI_END_STATE) {*/
AC3_AUDIO_MODULE_process_pkt_read_BSI:
            AC3_AUDIO_read_BSI(&(data->frame));
         }
            
         while ( (data->frame.payload_len16)&&
                 (data->frame.state >= AUDIO_BLOCK_read_config_STATE) ) {
AC3_AUDIO_MODULE_process_pkt_LOOP:
            if (data->frame.state < AUDIO_BLOCK_read_config_END_STATE)
               AUDIO_BLOCK_read_config(&(data->frame), &(data->blk));
            else if (data->frame.state < AUDIO_BLOCK_read_exp_END_STATE)
               AUDIO_BLOCK_read_exp(&(data->frame), &(data->blk));
            else if (data->frame.state < AUDIO_BLOCK_read_bit_alloc_END_STATE)
               AUDIO_BLOCK_read_bit_alloc(&(data->frame), &(data->blk));
            else {
               AUDIO_BLOCK_read_mant(&(data->frame), &(data->blk));
               if (data->blk.num > AC3_OUTPUT_BLK)
                  output_block(&(data->frame), &(data->blk), data->sound_init);
            }
         }
         data->frame.frame_len -= (u32)(pkt->payload_len16-data->frame.payload_len16);
#ifdef AC3_AUDIO_MODULE_TRACE
         printf("  state=0x%04hX  bytes=%u  left=%hu\n", 
                data->frame.state, data->frame.frame_len,
                data->frame.payload_len16);
         if (data->frame.state < AC3_AUDIO_read_BSI_STATE)
            printf("<END___FRAME>\n\n");
#endif
         pkt->payload_len16 = data->frame.payload_len16;
         pkt->payload       = data->frame.payload;
      }
   }
#ifdef AC3_AUDIO_MODULE_TRACE
   printf("end AC3_AUDIO_MODULE_process_pkt()\n\n");
#endif
}


static void init_audio_buf(AC3_SYNC_FRAME *frame, AUDIO_BLOCK *blk, u32 num_ch)
{
#if (AC3_OUTPUT_CHANNELS == 2)
   blk->audio_buf[1] = blk->audio_buf[0] + 1;
#else
   if (frame->mode & AC3_AUDIO_AC_LFE) {
      if (num_ch == 2)
         blk->audio_buf[frame->num_ch] = NULL;
      else {
         blk->audio_buf[frame->num_ch] = blk->audio_buf[0] +
                                         SOUND_51_SURROUND_get_pos(SUR51_CH_LFE);
      }
   }

   switch (AC3_AUDIO_AC(frame)) {
   case AC3_AUDIO_AC_10:
      break;

   case AC3_AUDIO_AC_1p1:  /* CH1 CH2 */
   case AC3_AUDIO_AC_20:   /* LFT RGT */
      blk->audio_buf[1] = blk->audio_buf[0];
      if (num_ch == 2)
         blk->audio_buf[1]++;
      else {
         blk->audio_buf[1] += SOUND_51_SURROUND_get_pos(SUR51_CH_FRGT);
         blk->audio_buf[0] += SOUND_51_SURROUND_get_pos(SUR51_CH_FLFT);
      }
      break;

   case AC3_AUDIO_AC_30:   /* LFT CEN RGT */
      blk->audio_buf[2] = 
      blk->audio_buf[1] = blk->audio_buf[0];
      if (num_ch == 2) {
         blk->audio_buf[2]++;
         blk->audio_buf[1]++;
      }
      else {
         blk->audio_buf[2] += SOUND_51_SURROUND_get_pos(SUR51_CH_FRGT);
         blk->audio_buf[1] += SOUND_51_SURROUND_get_pos(SUR51_CH_FCEN);
         blk->audio_buf[0] += SOUND_51_SURROUND_get_pos(SUR51_CH_FLFT);
      }
      break;

   case AC3_AUDIO_AC_21:   /* LFT RGT SUR */
      blk->audio_buf[2] = 
      blk->audio_buf[1] = blk->audio_buf[0];
      if (num_ch == 2) {
         blk->audio_buf[2] = NULL;
         blk->audio_buf[1]++;
      }
      else {
         /* TODO: Find a way to copy the single SUR channel to 
                  left & right SUR channels. */
         blk->audio_buf[2] += SOUND_51_SURROUND_get_pos(SUR51_CH_RLFT);
         blk->audio_buf[1] += SOUND_51_SURROUND_get_pos(SUR51_CH_FRGT);
         blk->audio_buf[0] += SOUND_51_SURROUND_get_pos(SUR51_CH_FLFT);
      }
      break;

   case AC3_AUDIO_AC_31:   /* LFT CEN RGT SUR */
      blk->audio_buf[3] = 
      blk->audio_buf[2] = 
      blk->audio_buf[1] = blk->audio_buf[0];
      if (num_ch == 2) {
         blk->audio_buf[3] = NULL;
         blk->audio_buf[2]++;
         blk->audio_buf[1]++;
      }
      else {
         /* TODO: Dind a way to copy the single SUR channel to 
                  left & right SUR channels. */
         blk->audio_buf[3] += SOUND_51_SURROUND_get_pos(SUR51_CH_RLFT);
         blk->audio_buf[2] += SOUND_51_SURROUND_get_pos(SUR51_CH_FRGT);
         blk->audio_buf[1] += SOUND_51_SURROUND_get_pos(SUR51_CH_FCEN);
         blk->audio_buf[0] += SOUND_51_SURROUND_get_pos(SUR51_CH_FLFT);
      }
      break;

   case AC3_AUDIO_AC_22:   /* LFT RGT LSUR RSUR */
      blk->audio_buf[3] = 
      blk->audio_buf[2] = 
      blk->audio_buf[1] = blk->audio_buf[0];
      if (num_ch == 2) {
         blk->audio_buf[2] = NULL;
         blk->audio_buf[1]++;
      }
      else {
         blk->audio_buf[3] += SOUND_51_SURROUND_get_pos(SUR51_CH_RRGT);
         blk->audio_buf[2] += SOUND_51_SURROUND_get_pos(SUR51_CH_RLFT);
         blk->audio_buf[1] += SOUND_51_SURROUND_get_pos(SUR51_CH_FRGT);
         blk->audio_buf[0] += SOUND_51_SURROUND_get_pos(SUR51_CH_FLFT);
      }
      break;

   case AC3_AUDIO_AC_32:   /* LFT CEN RGT LSUR RSUR */
      blk->audio_buf[4] = 
      blk->audio_buf[3] = 
      blk->audio_buf[2] = 
      blk->audio_buf[1] = blk->audio_buf[0];
      if (num_ch == 2) {
         blk->audio_buf[3] = NULL;
         blk->audio_buf[2]++;
         blk->audio_buf[1]++;
      }
      else {
         blk->audio_buf[4] += SOUND_51_SURROUND_get_pos(SUR51_CH_RRGT);
         blk->audio_buf[3] += SOUND_51_SURROUND_get_pos(SUR51_CH_RLFT);
         blk->audio_buf[2] += SOUND_51_SURROUND_get_pos(SUR51_CH_FRGT);
         blk->audio_buf[1] += SOUND_51_SURROUND_get_pos(SUR51_CH_FCEN);
         blk->audio_buf[0] += SOUND_51_SURROUND_get_pos(SUR51_CH_FLFT);
      }
      break;

   default:
      blk->audio_buf[1] = NULL;
   }
#endif
}

static void output_block(AC3_SYNC_FRAME *frame, AUDIO_BLOCK *blk, u32 num_ch)
{
   CH_STRUCT  *cur_ch     = blk->ch_data;
   s16       **audio      = blk->audio_buf,
              *delay      = blk->delay_buf;
   u8          blk_switch = blk->blk_switch;
#ifdef AC3_AUDIO_MODULE_TRACE
   int i = 0, j;
#endif
   blk->num -= AC3_OUTPUT_BLK;
#ifdef AC3_AUDIO_MODULE_TRACE
   printf("output_block(num=%u, blk_switch=0x%02hX",
          (blk->num-1), (u16)blk_switch);
   if (blk->num == 1)
      printf(", num_ch=%u)\n", num_ch);
   else
      printf(")\n");
#endif
   do {
#ifdef AC3_AUDIO_MODULE_TRACE
      printf("channel[%d]\ninput\n", i++);
      for (j = 0; j < AC3_NUM_BINS; j++) {
         printf("  [%3d]=%g\n", j, 
                (double)cur_ch->coeff[j]/(double)AC3_COEFF_CONST(1.0));
      }
#endif
      if (blk_switch >= 0x80) {
         AC3_imdft256(cur_ch->coeff);
         AC3_window256(cur_ch->coeff, *audio, delay, num_ch);
#ifdef AC3_AUDIO_MODULE_TRACE
         printf("2*256");
#endif
      }
      else {
         AC3_imdft512(cur_ch->coeff);
         AC3_window512(cur_ch->coeff, *audio, delay, num_ch);
#ifdef AC3_AUDIO_MODULE_TRACE
         printf("512");
#endif
      }
#ifdef AC3_AUDIO_MODULE_TRACE
      printf(" sample block output\n");
      for (j = 0; j < AC3_NUM_BINS; j++) {
         printf("  [%3d]=%6hd\n", j, 
                *audio[j*(int)num_ch]);
      }
      printf("\n");
#endif
      audio++;
      delay += AC3_NUM_BINS;
      blk_switch <<= 1;
   } while ( (++cur_ch < (CH_STRUCT *)blk->lfe_ptr)&&(*audio) );

#if (AC3_OUTPUT_CHANNELS > 2)
   if ( (blk->lfe_ptr->num_bins)&&(*audio) ) {
      cur_ch--;
      memcpy(cur_ch->coeff, blk->lfe_ptr->coeff, 
             (AC3_MAX_LFE_BIN*sizeof(AC3_COEFF)));
      memset((cur_ch->coeff+AC3_MAX_LFE_BIN), 0,
             ((AC3_NUM_BINS-AC3_MAX_LFE_BIN)*sizeof(AC3_COEFF)));
# ifdef AC3_AUDIO_MODULE_TRACE
      printf("channelLFE\ninput\n");
      for (j = 0; j < AC3_MAX_LFE_BIN; j++) {
         printf("  [%3d]=%g\n", j, 
                (double)cur_ch->coeff[j]/(double)AC3_COEFF_CONST(1.0));
      }
# endif
      AC3_imdft512(cur_ch->coeff);
      AC3_window512(cur_ch->coeff, *audio, delay, num_ch);
# ifdef AC3_AUDIO_MODULE_TRACE
      printf("512");
      printf(" sample block output\n");
      for (j = 0; j < AC3_NUM_BINS; j++) {
         printf("  [%3d]=%6hd\n", j,
                *audio[j*(int)num_ch]);
      }
      printf("\n");
# endif
   }
#endif

#ifndef AC3_AUDIO_MODULE_TRACE
# ifndef _SPEED_TEST_
   if (frame->pes_pts) {
      CLK_VAL cur_pts;

      cur_pts = MPEG2_prog_clk(frame->offset) +
                (CLK_VAL)((MPEG2_SYS_CLK_FREQ*SOUND_samples_left())/frame->sample_rate);
      *(frame->a_sync) = cur_pts - frame->pes_pts;
      frame->pes_pts   = 0;
   }
# endif
   SOUND_play();
#endif
}

