/***************************************************************************
                          PES_PKT.c  -  description
                             -------------------
   Stuff needed for handling PES packets.
                             -------------------
    begin                : Thu Sep 9 2004
    copyright            : (C) 2004 by Dennis Lew
    email                : dennis_L76@yahoo.ca
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "PES_PKT.h"


/* Reads the first PES packet header, in data_stream, into pes.
      INPUT: data_stream - Points to the data stream holding a PES packet.
             stream_len  - Number of bytes in the data stream.
      OUTPUT: pes - Points to the PES_PKT that will hold data taken from
                      data_stream.
      RETURN: Finished reading packet - Non-zero value.
              More data is needed     - 0 */
u32 PES_PKT_read(const u8 *data_stream, 
                 u32       stream_len, 
                 PES_PKT  *pes)
{
#define SKIP_HEADER_STATE    14
#define PAYLOAD_STATE        15
#ifndef MPEG2_SYS_CLK64
# define PTS_SHIFT            2
#else
# define PTS_SHIFT            0
#endif
   while (stream_len) {
      switch (pes->state) {
      case 0:  /* look for start code prefix */
         if (*data_stream == 0x00)
            pes->state++;
         break;
      case 1:
         if (*data_stream == 0x00)
            pes->state++;
         else
            pes->state = 0;
         break;
      case 2:
         if (*data_stream == 0x01)
            pes->state++;
         else
            pes->state = 0;
         break;

      case 3:  /* read stream id */
         pes->stream_id = *data_stream;
         pes->state++;
         break;
      
      case 4:  /* read packet length */
         pes->packet_length  = (u16)*data_stream << 8;
         pes->state++;
         break;
      case 5:
         pes->packet_length |= (u16)*data_stream;

         switch (pes->stream_id) {
         case PES_PROGRAM_STREAM_MAP:
         case PES_PADDING_STREAM:
         case PES_PRIVATE_STREAM_2:
         case PES_ECM_STREAM:
         case PES_EMM_STREAM:
         case PES_ISO_13818_1A_STREAM:
         case PES_ITUT_H2221_E:
         case PES_PROGRAM_STREAM_DIRECTORY:
            pes->header_length = 0;
            pes->state         = PAYLOAD_STATE;
            break;
         default:
            pes->state++;
         }
         break;

      case 6:  /* read PES header flags */
         pes->flags1 = *data_stream;
         pes->state++;
         break;
      case 7:
         pes->flags2 = *data_stream;
         pes->state++;
         break;

      case 8:  /* read header extension length */
         pes->header_length = *data_stream;
         if (PES_PTS_FLAG(pes))
            pes->state++;
         else if (pes->flags2 & 0x40)
            pes->state = PAYLOAD_STATE + 1;
         else
            pes->state = SKIP_HEADER_STATE;
         break;
         
      case 9:  /* read PTS */
         pes->PTS  = (CLK_VAL)(*data_stream & 0x0E) << (29 - PTS_SHIFT);
         pes->state++;
         break;
      case 10:
         pes->PTS |= (CLK_VAL) *data_stream         << (22 - PTS_SHIFT);
         pes->state++;
         break;
      case 11:
         pes->PTS |= (CLK_VAL)(*data_stream & 0xFE) << (14 - PTS_SHIFT);
         pes->state++;
         break;
      case 12:
         pes->PTS |= (CLK_VAL) *data_stream         << ( 7 - PTS_SHIFT);
         pes->state++;
         break;
      case 13:
         pes->PTS |= (CLK_VAL)(*data_stream         >> ( 1 + PTS_SHIFT));
#ifdef MPEG2_SYS_CLK64
         pes->PTS *= 300;
#endif
         if (pes->header_length <= 5)
            pes->state = PAYLOAD_STATE;
         else {
            pes->header_length -= 5;
            pes->state++;
         }
         break;

#ifdef DONOTCOMPILE         
         if (PES_DTS_FLAG(pes))
            pes->state++;
         else
            pes->state = SKIP_HEADER_STATE;
         break;

      case 14: /* read DTS */
         pes->DTS  = (s64)(*data_stream & 0x0E) << 29;
         pes->state++;
         break;
      case 15:
         pes->DTS |= (s64) *data_stream         << 22;
         pes->state++;
         break;
      case 16:
         pes->DTS |= (s64)(*data_stream & 0xFE) << 14;
         pes->state++;
         break;
      case 17:
         pes->DTS |= (s64) *data_stream         <<  7;
         pes->state++;
         break;
      case 18:
         pes->DTS |= (s64)(*data_stream         >>  1);
         pes->DTS *= 300;
         pes->header_length -= 5;
         pes->state++;
         break;
#endif
      
      case SKIP_HEADER_STATE: /* ignore bytes, before changing state */
         pes->header_length--;
         if (!pes->header_length)
            pes->state++;
         break;

      case PAYLOAD_STATE:     /* finally found the payload! */
         pes->state         = 0;
         pes->payload_len16 = (u16)stream_len;
         pes->payload       = data_stream;
         return 1;

      default:
         pes->state = 0;
         stream_len = 0;
      }
   
      if (stream_len) {
         stream_len--;
         data_stream++;
      }
      else
         break;
   }
#undef SKIP_HEADER_STATE
#undef PAYLOAD_STATE
#undef PTS_SHIFT
   return 0;
}

