/***************************************************************************
                          X11_display.c  -  description
                             -------------------
   Display module using the X11 library, XShm, and XVideo extensions.
                             -------------------
    begin                : Fri Nov 26 2004
    copyright            : (C) 2004 by Dennis Lew
    email                : dennis_L76@yahoo.ca
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
#define BACK_BUFFERS    2

#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <X11/Xlib.h>
#include <X11/extensions/XShm.h>
#include <X11/extensions/Xvlib.h>

#include "common/Timer.h"

static const char     *four_CC   = "YV12";
static u8             *ref_buf   = NULL;
static u32             buf_size  =  0,
                       displayed =  1;
static int             xv_check  =  1,
                       xv_port   = -1,
                       next_img  =  0;
static XvImage        *img[BACK_BUFFERS] = {NULL};
static XShmSegmentInfo shm_seg[BACK_BUFFERS];

static int  check_XVideo();
static int  init_XVideo(u32 src_wid, u32 src_hgt);
static void free_XVideo();

 
/* Prepares the display for the incoming 4:2:0 video stream.
      INPUT: src_wid  - Width of the source frame.
             src_hgt  - Height of the source frame.
             disp_wid - Width of the displayed frame.
             disp_hgt - Height of the displayed frame.
      RETURN: Success - Any non-zero value.
              Failure - 0 */
int DISPLAY_init(u32 src_wid, u32 src_hgt, u32 disp_wid, u32 disp_hgt)
{
   if (!init_X11(disp_wid, disp_hgt))
      return 0;
      
   if (xv_check) {
      if (!check_XVideo())
         goto DISPLAY_INIT_ERROR;   /* can not use XVideo */
      xv_check = 0;
   }
   
   if ( (!img[0])||
        (img[0]->width != src_wid)||(img[0]->height != src_hgt) ) {
      free_XVideo();
      if (!init_XVideo(src_wid, src_hgt))
         goto DISPLAY_INIT_ERROR;   /* can not use XVideo */
   }
   return 1;

DISPLAY_INIT_ERROR:
   DISPLAY_done();
   return 0;
}

/* Returns the buffer that will hold the reference frames. */
u8 *DISPLAY_ref_buf(u32 size)
{
   if (size > buf_size) {
      /* create a bigger buffer */
      if (ref_buf)
         free(ref_buf);
      ref_buf  = malloc((size_t)size);
      buf_size = (ref_buf) ? size : 0;
   }
   return ref_buf;
}

/* Called when the display is no longer needed. */
void DISPLAY_done()
{
   free_XVideo();
   free_X11();
   if (ref_buf) {
      free(ref_buf);
      ref_buf  = NULL;
      buf_size = 0;
   }
}

/* Returns the width of the frame, this module is expecting.
   This value may not be equal to src_wid, so do not ignore. */
u32 DISPLAY_frame_wid()
{
   if (img[0])
      return (u32)img[0]->width;
   return 0;
}

/* Returns a pointer to the next frame to be displayed.
   Returns NULL if DISPLAY_init was not called, or failed. */
u8 *DISPLAY_get_next_frame()
{
   if (img[next_img]) {
      displayed = 0;
      return (u8 *)img[next_img]->data;
   }
   return NULL;
}

/* Called when the next frame should be displayed.
   Returns a non-zero value, if the frame is late. */
u32 DISPLAY_next_frame_done(CLK_VAL pts, CLK_VAL offset)
{
#ifndef _SPEED_TEST_
   s32 sleep_time = (s32)(pts - MPEG2_prog_clk(offset));
#endif

#ifndef _SPEED_TEST_
   if (!displayed) {
# if (MPEG2_SYS_SLOW_DOWN > 1)
      sleep_time *= MPEG2_SYS_SLOW_DOWN;
# endif
      if (sleep_time > (MPEG2_SYS_CLK_FREQ / 1000)) {
         sleep_time /=  MPEG2_SYS_CLK_FREQ / 1000;
         TIMER_sleep(sleep_time);
      }
#endif
   
      if (img[next_img]) {
         XvShmPutImage(X_comm, xv_port, win, gc, img[next_img],
                       0, 0, img[next_img]->width, img[next_img]->height,
                       win_x, win_y, win_wid, win_hgt, False);
         XFlush(X_comm);
#if BACK_BUFFERS > 1
         next_img++;
         next_img %= BACK_BUFFERS;
#endif
#ifndef _SPEED_TEST_
         displayed = 1;
      }
   }
   return (sleep_time < -MAX_LAG_TIME);
#else
      }
   return 0;
#endif
}


/* Returns a non-zero value if XVideo is available for use. */
static int check_XVideo()
{
   u32                  ver, rel, req_base, evt_base, err_base;
   int                  retval, num_adaptors, num_formats, i, j, k;
   XvAdaptorInfo       *ai;
   XvImageFormatValues *ifv;

   xv_port = -1;
   if (!XShmQueryExtension(X_comm))
      return 0;

   retval = XvQueryExtension(X_comm, &ver, &rel,
                             &req_base, &evt_base, &err_base);
   if (retval != Success)
      return 0;

   retval = XvQueryAdaptors(X_comm, DefaultRootWindow(X_comm),
                            &num_adaptors, &ai);
   if ( (retval != Success)||(!num_adaptors) )
      return 0;

   for (i = 0; (i < num_adaptors)&&(xv_port < 0); i++) {
      if (!(ai[i].type & XvImageMask))
         continue;

      retval = ai[i].base_id + ai[i].num_ports;
      for (j = ai[i].base_id; (j < retval)&&(xv_port < 0); j++) {
         ifv = XvListImageFormats(X_comm, j, &num_formats);
         for (k = 0; k < num_formats; k++) {
            int *id = (int *)four_CC;

            if ( (ifv[k].type           == XvYUV)&&
                 (ifv[k].format         == XvPlanar)&&
                 (ifv[k].id             == *id)&&
                 (ifv[k].scanline_order == XvTopToBottom) ) {
               xv_port = j;   /* ai[i].base_id */
               break;
            }
         }
         if (num_formats)
            XFree(ifv);
      }
   }
   XvFreeAdaptorInfo(ai);

   if (xv_port < 0)
      return 0;
   return 1;
}

/* Returns a non-zero value when XVideo is ready for use. */
static int init_XVideo(u32 src_wid, u32 src_hgt)
{
   int  i,
       *id = (int *)four_CC;

   for (i = 0; i < BACK_BUFFERS; i++) {
      img[i] = XvShmCreateImage(X_comm, xv_port, *id, NULL,
                                src_wid, src_hgt, &(shm_seg[i]));
      shm_seg[i].shmid = shmget(IPC_PRIVATE, img[i]->data_size,
                                (IPC_CREAT | 0777));
      if (shm_seg[i].shmid < 0) {
         XFree(img[i]);
         img[i] = NULL;
         break;
      }

      shm_seg[i].shmaddr = img[i]->data = shmat(shm_seg[i].shmid, 0, 0);
      shm_seg[i].readOnly = False;
      if (!XShmAttach(X_comm, &(shm_seg[i]))) {
         shmdt(img[i]->data);
         shm_seg[i].shmaddr = img[i]->data = NULL;
         shmctl(shm_seg[i].shmid, IPC_RMID, NULL);
         XFree(img[i]);
         img[i] = NULL;
         break;
      }
   }

   if (i < BACK_BUFFERS) {
      while (i > 0) {
         i--;
         XShmDetach(X_comm, &(shm_seg[i]));
         shmdt(img[i]->data);
         shm_seg[i].shmaddr = img[i]->data = NULL;
         shmctl(shm_seg[i].shmid, IPC_RMID, NULL);
         XFree(img[i]);
         img[i] = NULL;
      }
      return 0;
   }
   return 1;
}

static void free_XVideo()
{
   if (img[0]) {
      int i;

      for (i = 0; i < BACK_BUFFERS; i++) {
         XShmDetach(X_comm, &(shm_seg[i]));
         shmdt(img[i]->data);
         shm_seg[i].shmaddr = img[i]->data = NULL;
         shmctl(shm_seg[i].shmid, IPC_RMID, NULL);
         XFree(img[i]);
         img[i] = NULL;
      }
   }
   else
      memset(img, 0, BACK_BUFFERS*sizeof(XvImage *));
}

#undef BACK_BUFFERS

