/***************************************************************************
                         mpeg_dvd.c  -  description
                             -------------------
   API to provide DVD access to this application.  This API mimics the
    one found in mpeg_ps.h
                             -------------------
    begin                : Mon Feb 6 2006
    copyright            : (C) 2006 by Dennis Lew
    email                : dennis_L76@yahoo.ca
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
#ifdef _DVD_DISPLAY_ERRORS_
# include <stdio.h>
#endif
#include <string.h>

#include "common/Timer.h"
#include "common/File.h"
#include "commands.h"
#include "AVindex.h"
#include "MPEG2_SYS.h"
#include "mpeg_ps.h"
#include "mpeg_ps_priv.h"
#include "mpeg_es.h"
#include "mpeg_dvd.h"

static void MPEG_DVD_jump(DVD_PTR        dvd, 
                          MPEG2_SYS_ARG *v_arg, 
                          MPEG2_SYS_ARG *a_arg);
static void MPEG_DVD_jump_chapter(DVD_PTR        dvd, 
                                  MPEG2_SYS_ARG *v_arg, 
                                  MPEG2_SYS_ARG *a_arg);


/* Gets information about the DVD.
      INPUT: dvd  - DVD pointer.
             size - Number of elements in the chapters array.
      OUTPUT: title    - Title of the DVD.
              chapters - Array of chapters, for each title, on the
                           DVD.
      RETURN: The number of titles found. */
u32 MPEG_DVD_get_info(DVD_PTR dvd, const char **title, u16 *chapters, u32 size)
{
   u32 num_titles, i, j;

   memset(chapters, 0, (sizeof(u16)*(size_t)size));

   if ( (dvdnav_get_title_string(dvd, title) == DVDNAV_STATUS_OK)&&
        (dvdnav_get_number_of_titles(dvd, &num_titles) == DVDNAV_STATUS_OK) ) {
      if (num_titles > size)
         num_titles = size;
      for (i = 0; i < num_titles; i++) {
         if (dvdnav_get_number_of_parts(dvd, (int32_t)(i+1), &j) != DVDNAV_STATUS_OK)
            break;
         chapters[i] = (u16)j;
      }
      return i;
   }
#ifdef _DVD_DISPLAY_ERRORS_
   fprintf(stderr, dvd_err_str, dvdnav_err_to_string(dvd));
#endif
   return 0;
}

/* Creates a program, using the first audio and video streams
     found in the DVD.
      INPUT: dvd - DVD pointer.
      RETURN: Program number, of newly created program, is stored
                in the global struct (sys.program_number).
              Success - A non-zero value
              Failure - 0 */
void MPEG_DVD_Create_Program(DVD_PTR dvd)
{
   DVD_POS           dvd_pos;
   int               mod;
   PS_CREATE_PROGRAM arg      = {TS_PKT_NUL_ID, TS_PKT_NUL_ID};
   u8                vid_type = 0,
                     aud_type = 0;

   DVD_GETPOS(dvd, dvd_pos);
   MPEG_DVD_decode(dvd, (dvd_pos+DVD_BYTES_TO_BLKS(CMD_SEARCH_SIZE)),
                   PS_create_program, &arg);
   DVD_SETPOS(dvd, dvd_pos);

   /* find correct stream type values */
   if (arg.vid_PID < TS_PKT_NUL_ID) {
      vid_type = MPEG_DVD_stream_type(dvd, arg.vid_PID);
      if (!vid_type)
         arg.vid_PID = TS_PKT_NUL_ID;
      else {
         mod = MPEG_ES_find_module(vid_type);
         if (mod < 0)
            arg.vid_PID = TS_PKT_NUL_ID;
         else
            vid_type    = (u8)mod;
      }
   }
   if (arg.aud_PID < TS_PKT_NUL_ID) {
      u16 org_pid = arg.aud_PID;
TRY_ORG_PID:
      if ((arg.aud_PID&0x00FF) == 0x00BD)
         arg.aud_PID = (DVD_AUDIO_DEFAULT(dvd) << 8) | 0xBD;
      else if ((arg.aud_PID&0xFFE0) == 0x00C0)
         arg.aud_PID =  DVD_AUDIO_DEFAULT(dvd)       | 0xC0;
      aud_type = MPEG_DVD_stream_type(dvd, arg.aud_PID);
      if (!aud_type) {
         if (arg.aud_PID != org_pid) {
            arg.aud_PID = org_pid;
            goto TRY_ORG_PID;
         }
         arg.aud_PID = TS_PKT_NUL_ID;
      }
      else {
         mod = MPEG_ES_find_module(aud_type);
         if (mod < 0)
            arg.aud_PID = TS_PKT_NUL_ID;
         else
            aud_type    = (u8)mod;
      }
   }

   if ( (arg.vid_PID >= TS_PKT_NUL_ID)&&
        (arg.aud_PID >= TS_PKT_NUL_ID) )
      sys.program_number = 0;
   else {
      sys.program_number = 1;
      
      if (arg.vid_PID < TS_PKT_NUL_ID)
         sys.PID[vid_type] = arg.vid_PID;

      if (arg.aud_PID < TS_PKT_NUL_ID)
         sys.PID[aud_type] = arg.aud_PID;
   }
}

/* Gets the stream ids and types from the DVD.
      INPUT: dvd  - DVD pointer.
             size - Number of elements in the PID and stream_type
                      arrays.
      OUTPUT: PID         - Array of stream ids found in the packet
                              stream.
              stream_type - Array of STREAM_TYPE_* values 
                              corresponding to the values in the
                              PID array.
      RETURN: The number of PIDs found. */
u32 MPEG_DVD_get_PIDs(DVD_PTR dvd, u16 *PID, u8 *stream_type, u32 size)
{
   u32         i;
   DVD_POS     dvd_pos;
   PS_GET_PIDS arg = {PID, size, 0};

   memset(PID, 0xFF, sizeof(u16)*(size_t)size);
   memset(stream_type, 0, sizeof(u8)*(size_t)size);

   /* get PIDs first */
   DVD_GETPOS(dvd, dvd_pos);
   MPEG_DVD_decode(dvd, (dvd_pos+DVD_BYTES_TO_BLKS(CMD_SEARCH_SIZE)),
                   PS_get_PIDs, &arg);
   DVD_SETPOS(dvd, dvd_pos);

   /* now get the stream_types */
   size = arg.found;
   for (i = 0; i < size; i++) {
      stream_type[i] = MPEG_DVD_stream_type(dvd, PID[i]);
   }
   return size;
}

/* Tries to determine the type of data contained in the stream with
     the given PID.
      INPUT: dvd - DVD pointer.
             PID - PID of the stream to look for.
      RETURN: Success - One of the STREAM_TYPE_* values defined
                          in PMT_SECTION.h
              Failure - Zero. */
u8 MPEG_DVD_stream_type(DVD_PTR dvd, u16 PID)
{
   PS_STREAM_TYPE arg;
   DVD_POS        dvd_pos;
   
   memset(&arg, 0, sizeof(PS_STREAM_TYPE));
   arg.PID = PID;

   DVD_GETPOS(dvd, dvd_pos);
   MPEG_DVD_decode(dvd, (dvd_pos+DVD_BYTES_TO_BLKS(CMD_SEARCH_SIZE)),
                   PS_stream_type, &arg);
   DVD_SETPOS(dvd, dvd_pos);

   return arg.mod_data.stream_type;
}

/* Decodes the MPEG program stream from the DVD.
      INPUT: dvd     - DVD pointer.
             end_pos - Point at which the decoding will stop. Decoding
                         will stop at an earlier point, if an EOF is
                         encountered.
             func    - Function that will deal with the decoded 
                         PES_PKTs.
             arg     - Pointer to any private data for func. */
void MPEG_DVD_decode(DVD_PTR             dvd, 
                     DVD_POS             end_pos, 
                     MPEG_PS_decode_func func,
                     void               *arg)
{
   const u8     *ptr;      
   DVD_POS       cur_pos;
   DVD_BLK_COUNT blks_read;
   u32           buf_size;
   PES_PKT       pes;

   end_pos = DVD_POS_BLOCKS(end_pos);
   cur_pos = 0;
   memset(&pes, 0, sizeof(PES_PKT));
   do {
      DVD_READ(dvd, sys.buf, DVD_BYTES_TO_BLKS(BUFFER_SIZE), blks_read);
      if (blks_read < 1)
         break;
      ptr      = sys.buf;
      buf_size = DVD_BLKS_TO_BYTES(blks_read);
      if (pes.payload) {
         pes.payload = ptr;
         if ((u32)pes.packet_length < buf_size) {
            pes.payload_len16 = pes.packet_length;
            ptr              += pes.packet_length;
         }
         else
            pes.payload_len16 = (u16)buf_size;
      }
      
      do {
         while ( (!pes.payload)&&(buf_size) ) {
            ptr = PS_find_packet(ptr, &buf_size, &pes);
         }

         if (pes.payload) {
            pes.packet_length -=      pes.payload_len16;
            buf_size          -= (u32)pes.payload_len16;

/*********** START TO DO SOMETHING WITH THE PES PACKET ***********/
            if (!func(&pes, arg)) {
               buf_size = blks_read = 0;
               break;
            }
/*********** STOP DOING SOMETHING WITH THE PES PACKET ************/

            pes.payload = (pes.packet_length) ? (const u8 *)1 : NULL;
         }
      } while (buf_size);
      
      cur_pos += (DVD_POS)blks_read;
   } while ( (blks_read == DVD_BYTES_TO_BLKS(BUFFER_SIZE))&&(cur_pos < end_pos) );
}

/* Plays the DVD.
      INPUT: dvd - DVD pointer.
      RETURN: Success - A non-zero value
              Failure - 0 */
int MPEG_DVD_Play(DVD_PTR dvd, u32 start_chapter)
{
#ifndef _SPEED_TEST_
   DVD_POS       dvd_pos, next_index_pt;
#endif
   DVD_BLK_COUNT blks_read;
   u8           *buf;
   u32           last_cmd;
#ifndef _NO_VIDEO_
   THREAD        v;
#endif
   THREAD        a;
   MPEG2_SYS_ARG v_arg = {NULL, 0, 0,              VIDEO_DECODERS, 0},
                 a_arg = {NULL, 0, VIDEO_DECODERS, MPEG2_SYS_PIDS, 0};
#ifdef _RECORD_STATS_
   TIME          t1, t2;
   
   sys.stream_bytes = sys.pkt_cnt = 0;
   sys.decode_time  = 0.0;
#endif
   if (sys.data.command == CMD_STOP)
      return 1;

   sys.data.a_pos  = sys.data.v_pos  =
   sys.data.a_skip = sys.data.v_skip = 0;

/*   printf("MAIN>  start\n");*/

   v_arg.buf = a_arg.buf = buf = sys.buf;
#ifndef _NO_VIDEO_
   v = new_THREAD(MPEG_PS_Play, &v_arg);
#endif
   a = new_THREAD(MPEG_PS_Play, &a_arg);
#ifndef _NO_VIDEO_
   if ( (v)&&(a) )
#else
   if (a)
#endif
   {
      MPEG2_SYS_modules_init();
      /* send initial offset value to all modules */
      sys.offset = MPEG2_get_offset(0);
      MPEG2_SYS_decoder_pause();

#ifndef _SPEED_TEST_
      DVD_GETPOS(dvd, dvd_pos);
      next_index_pt = dvd_pos + DVD_BYTES_TO_BLKS(AVINDEX_INTERVAL);
      AVindex_init((FILE_POS)dvd_pos, 0, 0, 0, 0);
#endif
      if (start_chapter > 1) {
         CLK_VAL offset   = sys.jump_offset;
         last_cmd         = sys.data.command;
         
         sys.jump_offset  = (CLK_VAL)(start_chapter-1);
         MPEG_DVD_jump_chapter(dvd, &v_arg, &a_arg);
         
         sys.jump_offset  = offset;
         sys.data.command = last_cmd;
      }
      if (sys.data.command == CMD_JUMP)
         MPEG_DVD_jump(dvd, &v_arg, &a_arg);
/*
      if (sys.data.command != CMD_STOP)
         printf("MAIN>  working\n");
*/
#ifdef _RECORD_STATS_
      TIMER_curtime(&t1);
#endif
      while (sys.data.command != CMD_STOP) {
         DVD_READ(dvd, buf, DVD_BYTES_TO_BLKS(BUFFER_SIZE), blks_read);
         if (blks_read < 1)
            break;

         lock_MUTEX(sys.m, 1);
#ifndef _NO_VIDEO_
         v_arg.bytes_read += (u32)DVD_BLKS_TO_BYTES(blks_read);
#endif
         a_arg.bytes_read += (u32)DVD_BLKS_TO_BYTES(blks_read);
         broadcast_CONDVAR(sys.cv);
         unlock_MUTEX(sys.m);

#ifdef _RECORD_STATS_
         sys.stream_bytes += (u32)DVD_BLKS_TO_BYTES(blks_read);
#endif
#ifndef _SPEED_TEST_
         DVD_GETPOS(dvd, dvd_pos);
#endif

         lock_MUTEX(sys.m, 1);
#ifndef _NO_VIDEO_
         while ( ( (sys.data.command != CMD_SLOW_FWD)&&
                   ( ( (v_arg.bytes_read)&&(a_arg.bytes_read) )||
                     (v_arg.bytes_read > (BUFFER_SIZE*(NUM_BUFFERS-2))+1)||
                     (a_arg.bytes_read > (BUFFER_SIZE*(NUM_BUFFERS-2))+1) ) )||
                 ( (sys.data.command == CMD_SLOW_FWD)&&
                   ( (v_arg.bytes_read > 1)||(a_arg.bytes_read > 1) ) ) )
#else
         while (a_arg.bytes_read > 1)
#endif
         {
            wait_CONDVAR(sys.cv, sys.m);
         }
#ifndef _NO_VIDEO_
         buf = (!v_arg.bytes_read) ? v_arg.buf : a_arg.buf;
#else
         buf = a_arg.buf;
#endif
#ifndef _SPEED_TEST_
         if ( (dvd_pos >= next_index_pt)
# ifndef _NO_VIDEO_
              &&(dvd_blk >= DVD_BYTES_TO_BLKS(v_arg.bytes_read))
              &&(dvd_blk >= DVD_BYTES_TO_BLKS(v_arg.bytes_read))
# endif
            ) {
            next_index_pt = dvd_pos + DVD_BYTES_TO_BLKS(AVINDEX_INTERVAL);
            AVindex_add((FILE_POS)dvd_pos, v_arg.pos, a_arg.pos, 
                        DVD_BYTES_TO_BLKS(v_arg.bytes_read),
                        DVD_BYTES_TO_BLKS(a_arg.bytes_read));
         }
#endif
         unlock_MUTEX(sys.m);

MPEG_DVD_Play_PAUSE_DONE:
         if (blks_read < DVD_BYTES_TO_BLKS(BUFFER_SIZE))
            break;
         else if (sys.data.command == CMD_PAUSE) {
#ifdef _RECORD_STATS_
            TIMER_curtime(&t2);
            TIMER_diff(&t1, &t2);
            sys.decode_time += TIMER_getdbl(&t1);
#endif
            MPEG2_SYS_decoder_pause(&sys);
#ifdef _RECORD_STATS_
            TIMER_curtime(&t1);
#endif
            goto MPEG_DVD_Play_PAUSE_DONE;
         }
         else if ( (sys.data.command & CMD_JUMP)
                   &&(sys.data.a_skip < 10)
# ifndef _NO_VIDEO_
                   &&(sys.data.v_skip < 10)
# endif
                 )
            MPEG_DVD_jump(dvd, &v_arg, &a_arg);
      }
#ifdef _RECORD_STATS_
      TIMER_curtime(&t2);
      TIMER_diff(&t1, &t2);
      sys.decode_time += TIMER_getdbl(&t1);
#endif
   }

   /* shutdown threads */
/*   printf("MAIN>  stop threads\n");*/
   last_cmd = sys.data.command;
   lock_MUTEX(sys.m, 1);
#ifndef _NO_VIDEO_
   while ( (v_arg.bytes_read > 1)||
           (a_arg.bytes_read > 1) )
#else
   while (a_arg.bytes_read > 1)
#endif
   {
      wait_CONDVAR(sys.cv, sys.m);  /* wait for threads to finish processing */
   }
   v_arg.bytes_read = a_arg.bytes_read = 2;
   sys.data.command = CMD_STOP;
   unlock_MUTEX(sys.m);
   
   lock_MUTEX(sys.m, 1);
#ifndef _NO_VIDEO_
   while ( (v_arg.bytes_read)||(a_arg.bytes_read) )
#else
   while (a_arg.bytes_read)
#endif
   {
      broadcast_CONDVAR(sys.cv);
      wait_CONDVAR(sys.cv, sys.m);  /* wait for threads to shutdown */
   }
   v_arg.bytes_read = a_arg.bytes_read = 0xFFFFFFFF;
   broadcast_CONDVAR(sys.cv);       /* acknowledge shutdown */
   unlock_MUTEX(sys.m);

#ifndef _NO_VIDEO_
   if (v) {
      join_THREAD(v, NULL);
      del_THREAD(v);
# ifdef _RECORD_STATS_
      sys.pkt_cnt = v_arg.pkt_cnt;
# endif
/*      printf("MAIN>  VIDEO done!\n");*/
   }
#endif
   if (a) {
      join_THREAD(a, NULL);
      del_THREAD(a);
#ifdef _RECORD_STATS_
      if (a_arg.pkt_cnt > sys.pkt_cnt)
         sys.pkt_cnt = a_arg.pkt_cnt;
#endif
/*      printf("MAIN>  AUDIO done!\n");*/
   }
   sys.data.command = last_cmd;

#ifdef _RECORD_STATS_
   TIMER_curtime(&t1);
#endif
   MPEG2_SYS_modules_done();
#ifdef _RECORD_STATS_
   TIMER_curtime(&t2);
   TIMER_diff(&t1, &t2);
   sys.decode_time += TIMER_getdbl(&t1);
#endif
/*   printf("MAIN>  end\n");*/
#ifndef _NO_VIDEO_
   return ( (v)&&(a) );
#else
   return (int)a;
#endif
}


static void MPEG_DVD_jump(DVD_PTR        dvd, 
                          MPEG2_SYS_ARG *v_arg, 
                          MPEG2_SYS_ARG *a_arg)
{
   FILE_POS f_pos;
   DVD_POS  cur_pos;
   CLK_VAL  v_pos, a_pos, inc;
   u32      v_bck, a_bck;

   if (sys.data.command == CMD_JUMP_CHAPTER) {
      MPEG_DVD_jump_chapter(dvd, v_arg, a_arg);
      return;
   }
   
   DVD_GETPOS(dvd, cur_pos);

#ifndef _NO_VIDEO_
   lock_MUTEX(sys.m, 1);
   while ( (v_arg->bytes_read > 1)||(a_arg->bytes_read > 1) ) {
      wait_CONDVAR(sys.cv, sys.m);  /* wait for threads to finish processing */
   }
#endif

   inc              = sys.jump_offset;
NEW_JUMP_REQUEST:
   sys.jump_offset  = 0;
   sys.data.command = CMD_PLAY;
   broadcast_CONDVAR(sys.cv);       /* acknowledge jump request */
   unlock_MUTEX(sys.m);

   if (!inc) {
      sys.data.a_skip = sys.data.v_skip = 0;
      goto RESET_DVDPOS;
   }

#ifndef _NO_VIDEO_
   a_pos = sys.data.a_skip = sys.data.a_pos + inc;
   v_pos = sys.data.v_skip = sys.data.v_pos + inc;

   if ( (!AVindex_get(&f_pos, &v_pos, &a_pos, &v_bck, &a_bck))||
#else
   a_pos = sys.data.a_skip = sys.data.a_pos + inc;
   v_pos = 0; v_bck = 0;

   if ( (!AVindex_get(&f_pos, NULL,   &a_pos, NULL,   &a_bck))||
#endif
        ( (inc > 0)&&((DVD_POS)f_pos < cur_pos) )||
        ( (inc < 0)&&((DVD_POS)f_pos > cur_pos) ) ) {
      /* location is not in the index, so use default seek method */
      if (inc > 0) {
         sys.data.a_skip = sys.data.v_skip = inc;
      }
      else {
RESET_DVDPOS:
         sys.data.a_pos  = sys.data.v_pos  = 0;
         DVD_REWIND(dvd);
#ifndef _NO_VIDEO_
         MPEG2_SYS_discontinuity_alert(v_arg, a_arg);
#else
         MPEG2_SYS_discontinuity_alert(NULL, NULL);
#endif
      }
   }
   else {
      /* location was in the index, so use quick seek method */
      DVD_SETPOS(dvd, (DVD_POS)f_pos);
      sys.data.a_pos = a_pos; sys.data.a_skip -= a_pos;
      sys.data.v_pos = v_pos; sys.data.v_skip -= v_pos;
#ifndef _NO_VIDEO_
      MPEG2_SYS_discontinuity_alert(v_arg, a_arg);
      if ( (v_bck)||(a_bck) ) {
         u8           *buf, *buf2;
         DVD_BLK_COUNT blks_read = DVD_BYTES_TO_BLKS(BUFFER_SIZE);

         /* allow the lagging thread to catch up to leading thread */
         if (v_bck) {
            buf = a_arg->buf;
            if (a_arg->buf == sys.last_buf)
               v_arg->buf = sys.buf;
            else
               v_arg->buf = a_arg->buf + BUFFER_SIZE;
         }
         else {
            buf = v_arg->buf;
            if (v_arg->buf == sys.last_buf)
               a_arg->buf = sys.buf;
            else
               a_arg->buf = v_arg->buf + BUFFER_SIZE;
         }

         do {
            lock_MUTEX(sys.m, 1);
            if (v_bck) {
               if (v_bck > DVD_BYTES_TO_BLKS(BUFFER_SIZE)) {
                  v_arg->bytes_read = BUFFER_SIZE;
                  v_bck            -= DVD_BYTES_TO_BLKS(BUFFER_SIZE);
               }
               else {
                  blks_read         = v_bck;
                  v_arg->bytes_read = DVD_BLKS_TO_BYTES(v_bck);
                  v_bck             = 0;
               }
               buf2 = v_arg->buf;
            }
            else {
               if (a_bck > DVD_BYTES_TO_BLKS(BUFFER_SIZE)) {
                  a_arg->bytes_read = BUFFER_SIZE;
                  a_bck            -= DVD_BYTES_TO_BLKS(BUFFER_SIZE);
               }
               else {
                  blks_read         = a_bck;
                  a_arg->bytes_read = DVD_BLKS_TO_BYTES(a_bck);
                  a_bck             = 0;
               }
               buf2 = a_arg->buf;
            }
            DVD_READ(dvd, buf2, blks_read, blks_read);
            broadcast_CONDVAR(sys.cv);
            do {
               wait_CONDVAR(sys.cv, sys.m);
            } while (v_arg->bytes_read > 1);
            unlock_MUTEX(sys.m);
         } while ( (v_bck != a_bck)&&(blks_read == DVD_BYTES_TO_BLKS(BUFFER_SIZE)) );
         v_arg->buf = a_arg->buf = buf;
      }
#else
      MPEG2_SYS_discontinuity_alert(NULL, NULL);
#endif
   }

   lock_MUTEX(sys.m, 1);
   if (sys.data.command == CMD_JUMP) {
      if (sys.jump_offset) {
         inc  =  sys.jump_offset;
#if defined(_NO_VIDEO_)
         inc += sys.data.a_skip;
#elif defined(_NO_AUDIO_)
         inc += sys.data.v_skip;
#else
         inc += (sys.data.a_skip > sys.data.v_skip) ?
                 sys.data.a_skip : sys.data.v_skip;
#endif
      }
      else
         inc  = 0;
      DVD_GETPOS(dvd, cur_pos);
      goto NEW_JUMP_REQUEST;
   }
#ifdef _NO_VIDEO_
   sys.data.v_pos  =
   sys.data.v_skip = 0;
#endif
   unlock_MUTEX(sys.m);
}

static void MPEG_DVD_jump_chapter(DVD_PTR        dvd, 
                                  MPEG2_SYS_ARG *v_arg, 
                                  MPEG2_SYS_ARG *a_arg)
{
   int new_chapter;

#ifndef _NO_VIDEO_
   lock_MUTEX(sys.m, 1);
   while ( (v_arg->bytes_read > 1)||(a_arg->bytes_read > 1) ) {
      wait_CONDVAR(sys.cv, sys.m);  /* wait for threads to finish processing */
   }
#endif

NEW_JUMP_CHAPTER_REQUEST:
   new_chapter      = (int)sys.jump_offset;
   sys.jump_offset  = 0;
   sys.data.command = CMD_PLAY;
   broadcast_CONDVAR(sys.cv);       /* acknowledge jump request */
   unlock_MUTEX(sys.m);

   new_chapter += (int)dvd_chapter;
   if ( (new_chapter > 0)&&(new_chapter <= DVD_NUM_CHAPTERS(dvd)) ) {
      DVD_POS       new_pos;
      DVD_BLK_COUNT blks_read;

      new_pos = DVD_POS_MAKE(new_chapter, 0);
      DVD_SETPOS(dvd, new_pos);
      if (new_chapter == 1) {
         sys.data.a_pos = sys.data.v_pos = 0;
      }
      else {
         do {
            DVD_READ(dvd, sys.buf, 1, blks_read);
         } while (!dvd_time);
         sys.data.a_pos = sys.data.v_pos = dvd_time;
      }
      sys.data.a_skip = sys.data.v_skip = 0;
#ifndef _NO_VIDEO_
      MPEG2_SYS_discontinuity_alert(v_arg, a_arg);
#else
      MPEG2_SYS_discontinuity_alert(NULL, NULL);
#endif
   }

   lock_MUTEX(sys.m, 1);
   if (sys.data.command == CMD_JUMP_CHAPTER)
      goto NEW_JUMP_CHAPTER_REQUEST;

#ifdef _NO_VIDEO_
   sys.data.v_pos = 0;
#endif
   unlock_MUTEX(sys.m);
}

