/***************************************************************************
                    MPEG1_video_module.c  -  description
                             -------------------
   MPEG1 video PES module functions.
                             -------------------
    begin                : Thu Sep 16 2004
    copyright            : (C) 2004 by Dennis Lew
    email                : dennis_L76@yahoo.ca
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
/*#define OUTPUT_FRAMES   4*/
/*#define DISPLAY_ORDER*/
#if defined(_SHOW_ERRORS_)||defined(_SPEED_TEST_)||defined(DISPLAY_ORDER)||defined(OUTPUT_FRAMES)
# include <stdio.h>
#endif
#include <stdlib.h>
#include <string.h>

#ifdef DISPLAY_ORDER
# include "common/Timer.h"
# define SLEEP_TIME  3000
#endif
#include "MPEG2_SYS.h"
#include "PES_PKT.h"
#include "commands.h"
#include "display/display.h"
#include "MPEG1_video_module.h"
#include "PIC_HEADER.h"
#include "SLICE.h"
#include "mpeg2_video/SEQ_HEADER.h"
#include "mpeg2_video/motion_vectors.h"

#ifdef _SPEED_TEST_
static int frame_cnt = 0;
#endif

/*#define FIRST_SEQ 1*/
#ifdef FIRST_SEQ
static int seq_cnt   = 0,
           frame_cnt = 0;
#endif

typedef struct {
   SLICE1      slice;
   SEQ_HEADER  seq_hdr;
   PIC_HEADER1 pic_hdr;
   u32         init, bytes_stored;
   u16         code, start;
   SYS_DATA   *sys_data;
} MPEG1_video_module_data;


static u32 look_for_next_start_code(const u8                      *data_stream,
                                          u32                      data_len,
                                          MPEG1_video_module_data *data);
static void init_display(MPEG1_video_module_data *data);
static void new_frame(MPEG1_video_module_data *data);


/*****************************************************************
   GENERAL MODULE FUNCTIONS
     These functions are used by all SYSTEM, PSI, and PES modules.
 *****************************************************************/
/* Returns the number of bytes of memory needed by the module. */
size_t MPEG1_VIDEO_MODULE_mem_needed()
{
   return sizeof(MPEG1_video_module_data);
}

/* Initializes the memory provided for the module.
      INPUT: void *     - Pointer to the memory provided for the module.
             SYS_DATA * - Pointer to system data. */
void MPEG1_VIDEO_MODULE_init(void *mem, SYS_DATA *sys_data)
{
   MPEG1_video_module_data *data = (MPEG1_video_module_data *)mem;

#ifdef FIRST_SEQ
   data->start               = FIRST_SEQ;
#else
   data->start               = 1;
#endif
   data->code                = 0x8001;
   data->bytes_stored        = 0;
   data->seq_hdr.header_size = 0;

   if (!data->init) {
      data->init          = 1;
      data->slice.seq_hdr = &(data->seq_hdr);
      data->slice.pic_hdr = &(data->pic_hdr);
/*      printf("data       =0x%08X\n"
             "data->slice=0x%08X\n"
             "slice->QFS =0x%08X\n",
             (u32)data,
             (u32)&(data->slice),
             (u32)&(data->slice.QFS));*/
   }

   if (sys_data)
      data->sys_data = sys_data;

   data->slice.ref_tempF  =
   data->slice.ref_tempB  = PIC_DISPLAYED;
   data->slice.last_frame = CODING_TYPE_I_FRAME;
   data->slice.cur_temp   = 0;
}

/* Clean up (DO NOT FREE) the memory provided for the module.
      INPUT: void * - Pointer to the memory provided for the module. */
void MPEG1_VIDEO_MODULE_done(void *mem)
{
   MPEG1_video_module_data *data = (MPEG1_video_module_data *)mem;

#ifdef FIRST_SEQ
   data->start               = FIRST_SEQ;
#else
   data->start               = 1;
#endif
   data->code                = 0x8001;
   data->bytes_stored        = 0;
   data->seq_hdr.header_size = 0;

   if (data->slice.ref_buf) {
#ifdef _SPEED_TEST_
      if (frame_cnt > 0) {
         printf("VIDEO: FRAMES=%d  TARGET_FRAME_RATE=", frame_cnt);
         frame_cnt = 0;
         switch (SEQ_HEADER_FRAME_RATE_CODE(data->slice.seq_hdr)) {
         case FRAME_RATE_23976:
            printf("23.976\n");
            break;
         case FRAME_RATE_24:
            printf("24\n");
            break;
         case FRAME_RATE_25:
         case FRAME_RATE_50:
            printf("25\n");
            break;
         case FRAME_RATE_2997:
         case FRAME_RATE_5994:
            printf("29.97\n");
            break;
         case FRAME_RATE_30:
         case FRAME_RATE_60:
         default:
            printf("30\n");
         };
      }
#endif
      if (data->sys_data->command == CMD_PLAY) {
#ifdef DISPLAY_ORDER
         printf("**FINAL FRAMES**\n");
#endif
         /* output reference frames that were not displayed */
         if (data->slice.last_frame == CODING_TYPE_B_FRAME)
            new_frame(data);
         data->pic_hdr.type     = CODING_TYPE_I_FRAME;
         data->slice.last_frame = CODING_TYPE_I_FRAME;
         if (data->slice.ref_tempF < PIC_DISPLAYED)
            data->slice.ref_tempF = 0;
         if (data->slice.ref_tempB < PIC_DISPLAYED)
            data->slice.ref_tempB = 0;
         while ( (!data->slice.ref_tempF)||(!data->slice.ref_tempB) ) {
            new_frame(data);
            data->slice.ref_tempB++;
         }
      }
      else if (data->sys_data->command == CMD_STOP) {
         DISPLAY_done();
         data->slice.ref_buf = NULL;
      }
      data->slice.ref_frameF = NULL;
      data->slice.ref_frameB = NULL;
   }
   data->slice.ref_tempF  =
   data->slice.ref_tempB  = PIC_DISPLAYED;
   data->slice.last_frame = CODING_TYPE_I_FRAME;
   data->slice.cur_temp   = 0;
}

/* Gives the module the offset value needed to obtain the program clock value.
      INPUT: void *  - Pointer to the memory provided for the module.
             CLK_VAL - offset value */
void MPEG1_VIDEO_MODULE_set_offset(void *mem, CLK_VAL offset)
{
   MPEG1_video_module_data *data = (MPEG1_video_module_data *)mem;

   if (data->sys_data->command != CMD_PAUSE)
      data->slice.cur_pts += offset - data->slice.offset;
   data->slice.offset = offset;
}

/* Alerts the module to a discontinuity in the data stream.
      INPUT: void * - Pointer to the memory provided for the module. */
void MPEG1_VIDEO_MODULE_discontinuity(void *mem)
{
   MPEG1_video_module_data *data = (MPEG1_video_module_data *)mem;

   if (data->seq_hdr.header_size) {
      if ( (data->sys_data)&&(data->sys_data->v_skip <= 0) )
         data->sys_data->v_skip = 1;
      data->code  = 0x80FF;
   }
   else {
      data->start = 1;
      data->code  = 0x8001;
   }
   data->bytes_stored     = 0;
   data->slice.ref_tempF  =
   data->slice.ref_tempB  = PIC_DISPLAYED;
   data->slice.last_frame = CODING_TYPE_I_FRAME;
   data->slice.cur_temp   = 0;
}


/**************************************************************************
   PES MODULE FUNCTIONS
     These functions are used by SYSTEM modules to process the  PES_PKTs
       they produce.
 **************************************************************************/
/* Initialize everything needed to handle the given elementary stream.
     The PES module will handle the elementary stream (if able) until the
     MPEG2_MODULE_done function is called, or this function is called
     again.
      INPUT: void *        - Pointer to the memory provided for the module.
             PMT_SECTION * - Pointer to the PMT data.
             int           - Index, in elem array, of the elementary stream
                               that needs to be handled.
      RETURN: Module is ready to handle
                the stream.               - Non-zero value.
              Module cannot handle the
                stream.                   - 0 */
int MPEG1_VIDEO_MODULE_stream_init(void *mem, PMT_SECTION *pmt, int idx)
{
/*   MPEG1_video_module_data *data = (MPEG1_video_module_data *)mem; */

   return (pmt->elem[idx]->stream_type == STREAM_TYPE_MPEG1_VIDEO);
}

/* Processes the packet handled by the module.
      INPUT: void * - Pointer to the memory provided for the module.
             void * - Pointer to the PES_PKT that needs handling.
                        Set pes->stream_id to zero to let the PES
                        module know to expect the start of a new
                        PES_PKT. */
void MPEG1_VIDEO_MODULE_process_pkt(void *mem, void *pes)
{
   u32                      bytes_read;
   MPEG1_video_module_data *data = (MPEG1_video_module_data *)mem;
   PES_PKT                 *pkt  = (PES_PKT *)pes;

#ifndef _SPEED_TEST_
   if (PES_PTS_FLAG(pkt))
      data->slice.pes_pts = (pkt->PTS) ? pkt->PTS : 1;
#endif

   while (pkt->payload_len16) {
      if (!data->code) {
         /* PIC_HEADER */
         bytes_read = PIC_HEADER1_read(pkt->payload,
                                       (u32)pkt->payload_len16,
                                       &(data->pic_hdr),
                                       &(data->bytes_stored),
                                       &(data->code));
         if (data->code >= 0x8000)
            new_frame(data);
      }
      else if (data->code <= 0x00AF) {
         /* SLICE */
         bytes_read = SLICE1_read(pkt->payload,
                                  (u32)pkt->payload_len16,
                                  &(data->slice),
                                  &(data->bytes_stored),
                                  &(data->code));
#ifdef FIRST_SEQ
         if (data->code == 0x0FFF) {
            printf("SLICE1_read ERROR (seq=%d, frame=%d)\n", 
                   seq_cnt, frame_cnt);
/*            fflush(stdin);
            getchar();*/
         }
#endif
      }
      else if (data->code >= 0x8000) {
         /* have not encountered the start code */
         bytes_read = look_for_next_start_code(pkt->payload,
                                               (u32)pkt->payload_len16,
                                               data);
      }
      else if (data->code == 0x00B8) {
         /* GOP_HEADER */
         bytes_read = GOP_HEADER1_read(pkt->payload,
                                       (u32)pkt->payload_len16,
                                       &(data->pic_hdr),
                                       &(data->bytes_stored),
                                       &(data->code));
         if (data->slice.ref_tempF < PIC_PREV_GROUP)
            data->slice.ref_tempF |= PIC_PREV_GROUP;
         if (data->slice.ref_tempB < PIC_PREV_GROUP)
            data->slice.ref_tempB |= PIC_PREV_GROUP;
#ifdef DISPLAY_ORDER
         if (data->code >= 0x8000)
            printf("**GOP**\n");
#endif
      }
      else if (data->code == 0x00B3) {
         /* SEQ_HEADER */
         bytes_read = SEQ_HEADER_read(pkt->payload,
                                      (u32)pkt->payload_len16,
                                      &(data->seq_hdr),
                                      &(data->bytes_stored),
                                      &(data->code));
         if ( (data->code >= 0x8000)&&(!data->slice.ref_frameF) )
            init_display(data);
#ifdef FIRST_SEQ
         if (data->code >= 0x8000) {
            seq_cnt++;
            frame_cnt = 0;
/*            printf("seq %d\n", seq_cnt);*/
         }
#endif
      }
      else if (data->code == 0x00B5) {
         /* EXT_DATA */
#ifdef _SHOW_ERRORS_
         printf("MPEG1_VIDEO_MODULE: EXT_DATA\n");
#endif
         goto MPEG1_VIDEO_MODULE_NOT_SUPPORTED;
      }
      else if (data->code == 0x00B2) {
         /* USR_DATA */
#ifdef _SHOW_ERRORS_
         printf("MPEG1_VIDEO_MODULE: USR_DATA\n");
#endif
MPEG1_VIDEO_MODULE_NOT_SUPPORTED:
         data->code |= 0x8000;
         continue;
      }
      else {
#ifdef _SHOW_ERRORS_
         if (data->code == 0x00B7) {
            /* SEQ_END_CODE */
            printf("MPEG1_VIDEO_MODULE: SEQ_END_CODE\n");
         }
         else
            printf("MPEG1_VIDEO_MODULE: Unknown code 0x%04X\n", data->code);
#endif
         MPEG1_VIDEO_MODULE_done(data);
         continue;
      }

#ifdef _SHOW_ERRORS_
      if (bytes_read > (u32)pkt->payload_len16) {
         /* This is here to make sure the functions used above
              are working properly. If you never see the following
              printf statement, then you can safely remove this code. */
         printf("code=0x%04hX, "
                "bytes_read=%u, "
                "payload_len16=%hu\n", data->code,
                                       bytes_read,
                                       pkt->payload_len16);
         break;
      }
#endif
      pkt->payload_len16 -= (u16)bytes_read;
      pkt->payload       += bytes_read;
   }
}


static u32 look_for_next_start_code(const u8                      *data_stream,
                                          u32                      data_len,
                                          MPEG1_video_module_data *data)
{
   const
   u8  *ptr = data_stream;
   u32  len = data_len;

   if (!data->bytes_stored) {
      while (1) {
START_SEARCH:
         ptr = memchr(ptr, 0, (size_t)len);
         if (!ptr)
            break;

         ptr++;
         len = data_len - (u32)(ptr - data_stream);
FOUND_FIRST_BYTE:
         if (len >= 3) {
            if ( (!ptr[0])&&(ptr[1] == 0x01) ) {
               ptr += 2;
               len -= 2;
               goto FOUND_START_CODE;
            }
         }
         else if (len == 2) {
            if ( (!ptr[0])&&(ptr[1] == 0x01) ) {
               data->bytes_stored = 3;
               break;
            }
         }
         else {
            if (!len)
               data->bytes_stored = 1;
            else if (!ptr[0])
               data->bytes_stored = 2;
            break;
         }
      }
   }
   else if (data->bytes_stored == 1) {
      data->bytes_stored = 0;
      goto FOUND_FIRST_BYTE;
   }
   else if (data->bytes_stored == 2) {
FOUND_FIRST_TWO_BYTES:
      if (*ptr == 0x01) {
         if (len >= 2) {
            ptr++;
            len--;
            goto FOUND_START_CODE;
         }
         data->bytes_stored = 3;
      }
      else if (!(*ptr)) {
         if (len >= 2) {
            ptr++;
            len--;
            goto FOUND_FIRST_TWO_BYTES;
         }
      }
      else {
         data->bytes_stored = 0;
         goto START_SEARCH;
      }
   }
   else {
FOUND_START_CODE:
      data->bytes_stored = 0;
      if (data->start) {
         if (*ptr != 0xB3)
#ifdef _SHOW_ERRORS_
            goto UNEXPECTED_CODE;
#else
            goto START_SEARCH;
#endif
#ifdef FIRST_SEQ
         if (data->start > 1) {
            data->start--;
            printf("skip seq %hu\n", (FIRST_SEQ - (u16)data->start));
            goto START_SEARCH;
         }
         else
#endif
            data->start = 0;
      }
      else {
         switch (data->code & 0x00FF) {
         case 0x0000:   /* PIC_HEADER */
            if ( (*ptr > 0xAF)&&(*ptr != 0xB5)&&(*ptr != 0xB2) )
#ifdef _SHOW_ERRORS_
               goto UNEXPECTED_CODE;
#else
               goto START_SEARCH;
#endif
            break;
         case 0x00B3:   /* SEQ_HEADER */
            if ( (*ptr != 0xB8)&&(*ptr != 0xB5)&&(*ptr != 0xB2) )
#ifdef _SHOW_ERRORS_
               goto UNEXPECTED_CODE;
#else
               goto START_SEARCH;
#endif
            break;
         case 0x00B2:   /* USR_DATA */
         case 0x00B5:   /* EXT_DATA */
            break;
         case 0x00B8:   /* GOP_HEADER */
            if ( (*ptr)&&(*ptr != 0xB5)&&(*ptr != 0xB2) )
#ifdef _SHOW_ERRORS_
               goto UNEXPECTED_CODE;
#else
               goto START_SEARCH;
#endif
            break;
         case 0x00FF:   /* end of picture */
            if (*ptr == 0xB7)
               MPEG1_VIDEO_MODULE_done(data);
            else if ( (*ptr)&&(*ptr != 0xB3)&&(*ptr != 0xB8) )
#ifdef _SHOW_ERRORS_
               goto UNEXPECTED_CODE;
#else
               goto START_SEARCH;
#endif
            break;
         default:       /* assume SLICE */
            if ( (*ptr > 0xAF)&&(*ptr != 0xB8)&&(*ptr != 0xB3)&&(*ptr != 0xB5)&&
                 (*ptr != 0xB2)&&(*ptr != 0xB7) )
#ifdef _SHOW_ERRORS_
               goto UNEXPECTED_CODE;
#else
               goto START_SEARCH;
#endif
         }
      }
      data->code = (u16)(*ptr);
      ptr++;
      return (u32)(ptr - data_stream);
   }
   return data_len;

#ifdef _SHOW_ERRORS_
UNEXPECTED_CODE:
# ifdef FIRST_SEQ
   if (!data->start)
# endif
      printf("MPEG1_VIDEO_MODULE ERROR: 0x%04hX -> 0x%02hX\n", data->code, (u16)*ptr);
   goto START_SEARCH;
#endif
}

static void init_display(MPEG1_video_module_data *data)
{
   u32 sar,
       disp_wid = (u32)data->seq_hdr.horiz_size,
       disp_hgt = (u32)data->seq_hdr.vert_size;
   
   switch (SEQ_HEADER1_ASPECT_RATIO(data->slice.seq_hdr)) {
   case ASPECT_RATIO_SAR_06735:
      sar =  6735;
      break;
   case ASPECT_RATIO_SAR_07031:
      sar =  7031;
      break;
   case ASPECT_RATIO_SAR_07615:
      sar =  7615;
      break;
   case ASPECT_RATIO_SAR_08055:
      sar =  8055;
      break;
   case ASPECT_RATIO_SAR_08437:
      sar =  8437;
      break;
   case ASPECT_RATIO_SAR_08935:
      sar =  8935;
      break;
   case ASPECT_RATIO_SAR_09157:
      sar =  9157;
      break;
   case ASPECT_RATIO_SAR_09815:
      sar =  9815;
      break;
   case ASPECT_RATIO_SAR_10255:
      sar = 10255;
      break;
   case ASPECT_RATIO_SAR_10695:
      sar = 10695;
      break;
   case ASPECT_RATIO_SAR_10950:
      sar = 10950;
      break;
   case ASPECT_RATIO_SAR_11575:
      sar = 11575;
      break;
   case ASPECT_RATIO_SAR_12015:
      sar = 12015;
      break;
   default:
      sar =     0;
   };

   if (sar) {
      disp_wid *= 10000; disp_hgt *=   sar;
      disp_wid /=   sar; disp_hgt /= 10000;
      
      if (disp_wid > (u32)data->seq_hdr.horiz_size)
         disp_hgt = (u32)data->seq_hdr.vert_size;
      else if (disp_hgt > (u32)data->seq_hdr.vert_size)
         disp_wid = (u32)data->seq_hdr.horiz_size;
   }
   
   data->slice.ref_tempF  = data->slice.ref_tempB = PIC_DISPLAYED+1;
   data->slice.cur_temp   = 0;
   data->slice.ref_frameF = NULL;
   data->slice.ref_frameB = NULL;

#ifndef _M1V_TRACE_
   if ( (DISPLAY_init((u32)data->seq_hdr.horiz_size,
                      (u32)data->seq_hdr.vert_size,
                      disp_wid, disp_hgt))&&
        (data->slice.frame_wid = DISPLAY_frame_wid()) ) {
      u32 bytes_needed = FRAME_420_SIZE(data->slice.frame_wid,
                                        (u32)data->seq_hdr.vert_size);
      
      bytes_needed      <<= 1;  /* we need 2 frames */
      data->slice.ref_buf = DISPLAY_ref_buf(bytes_needed);
      if (!data->slice.ref_buf) {
# ifdef _SHOW_ERRORS_
         printf("MPEG1_VIDEO_MODULE: "
                "Could not allocate memory for reference frames.\n");
# endif
         return;
      }
      data->slice.luma_size  = bytes_needed /  3;
      data->slice.frame_size = bytes_needed >> 1;
   }
   else {
      MPEG1_VIDEO_MODULE_done(data);
      data->code = 0x0FFF;
# ifdef _SHOW_ERRORS_
      printf("MPEG1_VIDEO_MODULE: "
             "Could not initialize display.\n");
# endif
   }
#else
   printf("Display Size = (%ux%u)\n\n", disp_wid, disp_hgt);
      data->slice.frame_wid  = (u32)data->seq_hdr.horiz_size;
      data->slice.frame_size = FRAME_420_SIZE((u32)data->seq_hdr.horiz_size,
                                              (u32)data->seq_hdr.vert_size);
      data->slice.luma_size  = (u32)data->seq_hdr.horiz_size *
                               (u32)data->seq_hdr.vert_size;
#endif

   data->slice.mb_wid = data->seq_hdr.horiz_size >> 4;
   data->slice.mb_hgt = data->seq_hdr.vert_size  >> 4;
   switch (SEQ_HEADER_FRAME_RATE_CODE(data->slice.seq_hdr)) {
   case FRAME_RATE_23976:
      data->seq_hdr.frame_interval = (s32)((100.1*MPEG2_SYS_CLK_FREQ) / 2400.0);
      break;
   case FRAME_RATE_24:
      data->seq_hdr.frame_interval = MPEG2_SYS_CLK_FREQ / 24;
      break;
   case FRAME_RATE_25:
   case FRAME_RATE_50:
      data->seq_hdr.frame_interval = MPEG2_SYS_CLK_FREQ / 25;
      break;
   case FRAME_RATE_2997:
   case FRAME_RATE_5994:
      data->seq_hdr.frame_interval = (s32)((100.1*MPEG2_SYS_CLK_FREQ) / 3000.0);
      break;
   case FRAME_RATE_30:
   case FRAME_RATE_60:
   default:
      data->seq_hdr.frame_interval = MPEG2_SYS_CLK_FREQ / 30;
   }
}

#ifdef OUTPUT_FRAMES
static void save_frame(u8 *frame, u32 luma_size, u32 num);
#endif

static void new_frame(MPEG1_video_module_data *data)
{
   SLICE1      *slice          = &(data->slice);
   PIC_HEADER1 *pic_hdr        = &(data->pic_hdr);
   SYS_DATA    *sys_data       = data->sys_data;
   s32          frame_interval = data->seq_hdr.frame_interval;
   
   if (!slice->ref_frameF) {
      if (PIC_HEADER1_CODING_TYPE(pic_hdr) != CODING_TYPE_I_FRAME) {
         data->code = 0x80FF; /* start of sequence must be I frame */
         return;
      }
      /* this is the first slice of the sequence */
#ifdef _M1V_TRACE_
      slice->ref_frameF =                     (u8 *)10000000;
      slice->ref_frameB = slice->pred_frame = (u8 *)20000000;
#else
      slice->ref_frameF = slice->ref_buf;
      slice->ref_frameB = slice->pred_frame = slice->ref_buf+slice->frame_size;
#endif
      slice->last_frame = CODING_TYPE_I_FRAME;
      slice->ref_tempB  = (u32)pic_hdr->temporal_reference;

      slice->cur_pts  = MPEG2_prog_clk(slice->offset) + (CLK_VAL)frame_interval;
      slice->ref_ptsF = 0;
      slice->ref_ptsB = slice->pes_pts;
#ifdef DISPLAY_ORDER
      printf("TEMP %2u| OUT? ??| IN_I %2u| DEC\n", 
             slice->cur_temp, slice->ref_tempB);
      TIMER_sleep(SLEEP_TIME);
#endif
   }
   else {
#if defined(FIRST_SEQ) || defined(_SPEED_TEST_)
      frame_cnt++;
#endif
#ifdef DISPLAY_ORDER
      printf("TEMP %2u| OUT", slice->cur_temp);
#endif
      /* this is the start of a new frame */
      if (slice->last_frame != CODING_TYPE_B_FRAME) {
         u8      *ref_frame;
         CLK_VAL  ref_pts;
#ifdef OUTPUT_FRAMES
         u32 num;
#endif

         if ( (!slice->cur_temp)&&
              (PIC_HEADER1_CODING_TYPE(pic_hdr) == CODING_TYPE_B_FRAME) )
            slice->cur_temp = (u32)pic_hdr->temporal_reference;

         if ( (slice->ref_tempF <= slice->cur_temp)||
              (slice->ref_tempF & PIC_PREV_GROUP) ) {
#ifdef OUTPUT_FRAMES
            num = (slice->ref_tempF & 0x0000FFFF);
#endif
#ifdef DISPLAY_ORDER
            printf("1 %2u|", (slice->ref_tempF & 0x0000FFFF)); 
#endif
            slice->ref_tempF = PIC_DISPLAYED;
            ref_frame        = slice->ref_frameF;
            ref_pts          = slice->ref_ptsF;
         }
         else if (!slice->ref_tempB) {
#ifdef OUTPUT_FRAMES
            num = 0;
#endif
#ifdef DISPLAY_ORDER
            printf("2  0|");
#endif
            slice->ref_tempB = PIC_DISPLAYED;
            ref_frame        = slice->ref_frameB;
            ref_pts          = slice->ref_ptsB;
         }
         else {
#ifdef DISPLAY_ORDER
            printf("? ??|");
#endif
            goto DISPLAY_FRAME;
         }

         AV_SYNC(slice->cur_pts, ref_pts, sys_data->a_sync)
         slice->cur_temp++;
#ifndef _M1V_TRACE_
# ifdef OUTPUT_FRAMES
         save_frame(ref_frame, slice->luma_size, num);
# endif
         MV_COPY_FRAME(DISPLAY_get_next_frame(), 
                       ref_frame, 
                       slice->frame_size);
#endif
      }
#if defined(OUTPUT_FRAMES)
      else
         save_frame(slice->pred_frame, slice->luma_size, (slice->cur_temp-1));
#elif defined(DISPLAY_ORDER)
      else
         printf("  %2u|", (slice->cur_temp-1));
#endif
DISPLAY_FRAME:
#ifndef _M1V_TRACE_
      /* output the next frame */
      if ( (DISPLAY_next_frame_done(slice->cur_pts, slice->offset))&&
           (sys_data->a_sync)&&(sys_data->v_skip <= 0) )
         sys_data->v_skip = 1;   /* drop frames, since we are behind */
#endif
#ifdef DISPLAY_ORDER
      printf(" IN_");
#endif
      /* store new frame type, and get next frame */
      slice->last_frame = PIC_HEADER1_CODING_TYPE(pic_hdr);
      if (slice->last_frame != CODING_TYPE_B_FRAME) {
         /* ref_frameB becomes the new ref_frameF */
         slice->ref_tempF  = slice->ref_tempB;
         slice->ref_tempB  = (u32)pic_hdr->temporal_reference;
         slice->pred_frame = slice->ref_frameF;
         slice->ref_frameF = slice->ref_frameB;
         slice->ref_frameB = slice->pred_frame;

         slice->ref_ptsF = slice->ref_ptsB;
         slice->ref_ptsB = slice->pes_pts;
#ifdef DISPLAY_ORDER
         printf("%c %2u|", 
                (slice->last_frame == CODING_TYPE_I_FRAME) ? 'I' : 'P',
                 slice->ref_tempB);
#endif
      }
      else {
         slice->cur_temp = (u32)pic_hdr->temporal_reference + 1;
#ifdef DISPLAY_ORDER
         printf("B %2u|", (slice->cur_temp-1));
#endif
         if (sys_data->v_skip <= 0) {
            if ( (slice->ref_tempF > PIC_DISPLAYED)||
                 (slice->ref_tempB > PIC_DISPLAYED) ) {
               slice->last_frame = 0;
               data->code        = 0x80FF;   /* wait for next picture */
# ifdef DISPLAY_ORDER
               printf(" SKIP_B\n");
               TIMER_sleep(SLEEP_TIME);
# endif
            }
            else {
#ifndef _M1V_TRACE_
               slice->pred_frame = DISPLAY_get_next_frame();
#else
               slice->pred_frame = (u8 *)30000000;
#endif
               AV_SYNC(slice->cur_pts, slice->pes_pts, sys_data->a_sync)
            }
         }
      }

#ifndef _SPEED_TEST_
      slice->pes_pts   = 0;
      sys_data->v_pos += (CLK_VAL)frame_interval;
      if (sys_data->v_skip > 0) {
         sys_data->v_skip -= (CLK_VAL)frame_interval;
         if ( (sys_data->v_skip <= 0)&&
              (slice->last_frame != CODING_TYPE_I_FRAME) )
            sys_data->v_skip = 1;   /* wait for an I frame */
         if (sys_data->v_skip > 0) {
            if (slice->last_frame != CODING_TYPE_B_FRAME)
               slice->ref_tempB = PIC_DISPLAYED+1;
            slice->last_frame = 0;
            data->code        = 0x80FF;   /* wait for next picture */
# ifdef DISPLAY_ORDER
            printf(" SKIP\n");
            TIMER_sleep(SLEEP_TIME);
# endif
            if (sys_data->v_skip > 1)
               return;
         }
         else
            slice->cur_pts = MPEG2_prog_clk(slice->offset);
      }

# ifdef DISPLAY_ORDER
      if (slice->last_frame) {
         printf(" DEC\n");
         TIMER_sleep(SLEEP_TIME);
      }
# endif
      if (sys_data->command == CMD_SLOW_FWD) {
         slice->pes_pts  = slice->ref_ptsF = slice->ref_ptsB = 0;
         slice->cur_pts += (CLK_VAL)(frame_interval << CMD_SLOW_SHIFT);
      }
      else
         slice->cur_pts += (CLK_VAL)frame_interval;
#endif
   }

   /* reset blk pointers */
   slice->ref_blkF = slice->ref_frameF;
   slice->ref_blkB = slice->ref_frameB;
   slice->pred_blk = slice->pred_frame;
      
   slice->cur_row = slice->cur_col = 0;
}

#ifdef OUTPUT_FRAMES
static int saved_frames = 0;
static void save_frame(u8 *frame, u32 luma_size, u32 num)
{

   char filename[64];
   FILE *fp;

   if (saved_frames >= OUTPUT_FRAMES)
      return;
   saved_frames++;

   sprintf(filename, "test/myframe%u.Y", num);
   fp = fopen(filename, "wb");
   if (!fp)
      return;
   fwrite(frame, 1, (size_t)luma_size, fp);
   fclose(fp);

   sprintf(filename, "test/myframe%u.V", num);
   fp = fopen(filename, "wb");
   if (!fp)
      return;
   frame += luma_size;
   fwrite(frame, 1, (size_t)(luma_size>>2), fp);
   fclose(fp);

   sprintf(filename, "test/myframe%u.U", num);
   fp = fopen(filename, "wb");
   if (!fp)
      return;
   frame += luma_size>>2;
   fwrite(frame, 1, (size_t)(luma_size>>2), fp);
   fclose(fp);

}
#endif

