/***************************************************************************
                          motion_vectors.c  -  description
                             -------------------
   These are the routines needed to handle motion vectors.
                             -------------------
    begin                : Wed Nov 24 2004
    copyright            : (C) 2004 by Dennis Lew
    email                : dennis_L76@yahoo.ca
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
#include "motion_vectors.h"

#if defined(__MMX__)

# if defined(__GCC__)
#  include "mmx/mmx_gcc_motion_vectors.c"
# elif defined(__MSVC__)
#  include "mmx/mmx_msvc_motion_vectors.c"
# else
#  error "Don't know how to insert inline ASM code for this compiler"
# endif

#else

# include <string.h>

/* Copy a frame. */
void MV_COPY_FRAME(void *to, void *from, u32 size)
{
   memcpy(to, from, (size_t)size);
}

/* Clear a block. */
void MV_CLEAR_BLOCK(s16 *blk)
{
   memset(blk, 0, 64*sizeof(s16));
}

   
/* The MV_PRED_?? functions are used to get the predictors from
     the reference block(s). */
void MV_PRED___(s16 *inp, u8 *ref, u32 inc)
{
   int i, j;
   
   inc -= 8;
   for (i = 0; i < 8; i++) {
      for (j = 0; j < 8; j++) {
         *inp += (s16)(*ref);
         ref++;
         inp++;
      }
      ref += inc;
   }   
}

void MV_PRED_X_(s16 *inp, u8 *ref, u32 inc)
{
   int i, j;
   s16 pred;
   
   inc -= 8;
   for (i = 0; i < 8; i++) {
      for (j = 0; j < 8; j++) {
         pred  = (s16)(*ref);
         ref++;
         pred += (s16)(*ref);
         *inp += pred >> 1;
         inp++;
      }
      ref += inc;
   }
}

void MV_PRED__Y(s16 *inp, u8 *ref, u32 inc, u32 inc_y)
{
   int i, j, pred;
   u8 *ptr0 = ref,
      *ptr1 = ref + inc_y;

   inc -= 8;
   for (i = 0; i < 8; i++) {
      for (j = 0; j < 8; j++) {
         pred  = (int)(*ptr0) + (int)(*ptr1);
         ptr0++;
         ptr1++;
         *inp += (s16)pred >> 1;
         inp++;
      }
      ptr0 += inc;
      ptr1 += inc;
   }
}

void MV_PRED_XY(s16 *inp, u8 *ref, u32 inc, u32 inc_y)
{
   int i, j, pred;
   u8 *ptr0 = ref,
      *ptr1 = ref + inc_y;

   inc -= 8;
   for (i = 0; i < 8; i++) {
      for (j = 0; j < 8; j++) {
         pred  = (int)(*ptr0) + (int)(*ptr1);
         ptr0++;
         ptr1++;
         pred += (int)(*ptr0) + (int)(*ptr1);
         *inp += (s16)pred >> 2;
         inp++;
      }
      ptr0 += inc;
      ptr1 += inc;
   }
}


/* The MV_FIELD_PRED_?? functions are similar to the MV_PRED_?? functions,
     except these functions are used when different reference blocks are
     needed for each field. */
void MV_FIELD_PRED___(s16 *inp0, u32 inp_inc, 
                      u8  *ref0, u32 inc)
{
   int i, j;
   
   inp_inc -= 8;
   inc     -= 8;
   for (i = 0; i < 4; i++) {
      for (j = 0; j < 8; j++) {
         *inp0 += (s16)(*ref0);
         ref0++;
         inp0++;
      }
      inp0 += inp_inc;
      ref0 += inc;
   }
}

void MV_FIELD_PRED_X_(s16 *inp0, u32 inp_inc, 
                      u8  *ref0, u32 inc)
{
   int i, j;
   s16 pred;
   
   inp_inc -= 8;
   inc     -= 8;
   for (i = 0; i < 4; i++) {
      for (j = 0; j < 8; j++) {
         pred   = (s16)(*ref0);
         ref0++;
         pred  += (s16)(*ref0);
         *inp0 += pred >> 1;
         inp0++;
      }
      inp0 += inp_inc;
      ref0 += inc;
   } 
}

void MV_FIELD_PRED__Y(s16 *inp0, u32 inp_inc, 
                      u8  *ref0, u32 inc, u32 inc_y)
{
   int i, j, pred;
   u8 *ref1 = ref0 + inc_y;
   
   inp_inc -= 8;
   inc     -= 8;
   for (i = 0; i < 4; i++) {
      for (j = 0; j < 8; j++) {
         pred   = (int)(*ref0) + (int)(*ref1);
         ref0++;
         ref1++;
         *inp0 += (s16)pred >> 1;
         inp0++;
      }
      inp0 += inp_inc;
      ref0 += inc;
      ref1 += inc;
   }
}

void MV_FIELD_PRED_XY(s16 *inp0, u32 inp_inc, 
                      u8  *ref0, u32 inc, u32 inc_y)
{
   int i, j, pred;
   u8 *ref1 = ref0 + inc_y;

   inp_inc -= 8;
   inc     -= 8;
   for (i = 0; i < 4; i++) {
      for (j = 0; j < 8; j++) {
         pred  = (int)(*ref0) + (int)(*ref1);
         ref0++;
         ref1++;
         pred += (int)(*ref0) + (int)(*ref1);
         *inp0 += (s16)pred >> 2;
         inp0++;
      }
      inp0 += inp_inc;
      ref0 += inc;
      ref1 += inc;
   }
}


/* Saturates the final pixel values, in the block, and stores them in
     the frame buffer. */
void MV_SATURATION(s16 *blk, u8 *outp, int inc)
{
   int i, j;
   
   inc -= 8;
   for (i = 0; i < 8; i++) {
      for (j = 0; j < 8; j++) {
         if (*blk <= -128)
            *outp = 0;
         else if (*blk >= 127)
            *outp = 255;
         else
            *outp = (u8)(*blk + 128);
         blk++;
         outp++;
      }
      outp += inc;
   }
}

   
/* Adds the predictor block to the DCT block, to obtain the final pixel
     values.  The final values are stored in the frame buffer. */
void MV_ADD_PRED(s16 *blk, s16 *p_blk, u8 *outp, int inc, u32 sh)
{
   int i, j;
   
   inc -= 8;
   for (i = 0; i < 8; i++) {
      for (j = 0; j < 8; j++) {
         *p_blk >>= sh;
         *p_blk  += *blk;
         if (*p_blk < 0)
            *outp = 0;
         else if (*p_blk > 255)
            *outp = 255;
         else
            *outp = (u8)(*p_blk);
         blk++;
         p_blk++;
         outp++;
      }
      outp += inc;
   }
}


/* Copies the predictor block into the frame buffer. */
void MV_COPY_PRED(s16 *p_blk, u8 *outp, int inc, u32 sh)
{
   int  i, j;
   
   inc -= 8;
   for (i = 0; i < 8; i++) {
      for (j = 0; j < 8; j++) {
         *p_blk >>= sh;
         if (*p_blk < 0)
            *outp = 0;
         else if (*p_blk > 255)
            *outp = 255;
         else
            *outp = (u8)(*p_blk);
         p_blk++;
         outp++;
      }
      outp += inc;
   }
}
   
#endif

