/***************************************************************************
                       cmdline_args.c  -  description
                             -------------------
   Stuff needed to deal with the command line arguments.
                             -------------------
    begin                : Fri Jul 16 2004
    copyright            : (C) 2004 by Dennis Lew
    email                : dennis_L76@yahoo.ca
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "cmdline_args.h"

#ifndef _NO_AUDIO_
# ifndef _NO_51_SURROUND_
#  include "sound/sound.h"
# endif

# ifdef __OSS_SOUND__
extern const char *OSS_DEVICE;
# endif

# ifdef __ALSA_SOUND__
extern const char *ALSA_DEVICE;
# endif
#endif


/* Parse the given command line arguments, starting with argv[idx].
   Returns: one of the CMDLINE_* values, or a positive number
            indicating the next argv[] to be parsed
 */
int CMDLINE_parse(int idx, int argc, char *argv[], APP_PARAMS *params)
{
   int rc;
   
   if (idx < 1)
      idx = 1;
   if (idx >= argc)
      return (argc <= 1) ? CMDLINE_USAGE : CMDLINE_DONE; /* nothing left to parse */
   else if (params == NULL) {
      fprintf(stderr, "CMDLINE ERROR: Missing APP_PARAMS structure!\n\n");
      return CMDLINE_ERROR;                              /* no parameters to set */
   }

   do {
      if (argv[idx][0] == '-') {
         if ( (!strcmp(argv[idx], "-h"))||(!strcmp(argv[idx], "-help")) )
            return CMDLINE_USAGE;
         else if (!strcmp(argv[idx], "-loop"))
            params->loop = 1;
         else if (!strcmp(argv[idx], "-pids"))
            params->req = DISPLAY_PIDS;
#ifdef MPEG2_SYS_CLK64
         else if (!strcmp(argv[idx], "-rates"))
            params->req = DISPLAY_RATES;
#endif
         else if (!strcmp(argv[idx], "-pat"))
            params->req = DISPLAY_PAT;
         else if (!strncmp(argv[idx], "-pmt=", 5)) {
            params->req = DISPLAY_PMT;
            rc          = sscanf((argv[idx] + 5), "%hx", &(params->prog_num));
            if ( (!rc)||(params->prog_num < 1) ) {
               fprintf(stderr, "\nERROR: Bad option %s\n"
                               "         Specify a value between"
                                        " 0x0001 and 0xFFFF\n", argv[idx]);
               return CMDLINE_ERROR;
            }
         }
         else if (!strcmp(argv[idx], "-pmt")) {
            fprintf(stderr, "\nERROR: -pmt option must have an argument\n"
                               "         Use -pmt=[n]\n");
            return CMDLINE_ERROR;
         }
         else if (!strncmp(argv[idx], "-play_prog=", 11)) {
            params->req = PLAY_PROG;
            rc          = sscanf((argv[idx] + 11), "%hx", &(params->prog_num));
            if ( (!rc)||(params->prog_num < 1) ) {
               fprintf(stderr, "\nERROR: Bad option %s\n"
                               "         Specify a value between"
                                        " 0x0001 and 0xFFFF\n", argv[idx]);
               return CMDLINE_ERROR;
            }
         }
         else if (!strcmp(argv[idx], "-play_prog")) {
            fprintf(stderr, "\nERROR: -play_prog option must have an argument\n"
                               "         Use -play_prog=[n]\n");
            return CMDLINE_ERROR;
         }
         else if (!strncmp(argv[idx], "-play_streams=", 14)) {
            params->req     = PLAY_STREAMS;
            params->PCR_PID = params->video_PID = params->audio_PID = 0xFFFF;
            rc              = sscanf((argv[idx] + 14),
                                     "%hx,%hx",
                                     &(params->video_PID),
                                     &(params->audio_PID));
            if (!rc) {
               fprintf(stderr, "\nERROR: Bad option %s\n"
                               "         Specify 1 or 2 PID values\n", argv[idx]);
               return CMDLINE_ERROR;
            }
         }
         else if (!strcmp(argv[idx], "-play_streams")) {
            fprintf(stderr, "\nERROR: -play_streams option must at least 1 argument\n"
                               "         Use -play_streams=[v],[a]\n");
            return CMDLINE_ERROR;
         }
         else if (!strncmp(argv[idx], "-start=", 7)) {
            double v1 = 0.0,
                   v2 = 0.0;
            params->start_time = 0.0;
            rc = sscanf((argv[idx] + 7), "%lf:%lf:%lf",
                        &(params->start_time), &v1, &v2);
            if (!rc) {
               fprintf(stderr, "\nERROR: Bad option %s\n"
                               "         Time format is H:M:S\n", argv[idx]);
               return CMDLINE_ERROR;
            }
            else if (rc == 3) {
               params->start_time *= 3600.0;
               params->start_time +=   60.0 * v1;
               params->start_time +=          v2;
            }
            else if (rc == 2) {
               params->start_time *=   60.0;
               params->start_time +=          v1;
            }
         }
         else if (!strcmp(argv[idx], "-start")) {
            fprintf(stderr, "\nERROR: -start option must have an argument\n"
                               "         Use -start=[time] where format is H:M:S\n");
            return CMDLINE_ERROR;
         }
#ifndef _NO_AUDIO_
# ifdef __OSS_SOUND__
         else if (!strncmp(argv[idx], "-oss_device=", 12))
            OSS_DEVICE = argv[idx] + 12;
         else if (!strcmp(argv[idx], "-oss_device")) {
            fprintf(stderr, "\nERROR: -oss_device option is missing the path to the device\n");
            return CMDLINE_ERROR;
         }
# endif
# ifdef __ALSA_SOUND__
         else if (!strncmp(argv[idx], "-alsa_device=", 13))
            ALSA_DEVICE = argv[idx] + 13;
         else if (!strcmp(argv[idx], "-alsa_device")) {
            fprintf(stderr, "\nERROR: -alsa_device option is missing the name of the device\n");
            return CMDLINE_ERROR;
         }
# endif
# ifndef _NO_51_SURROUND_
         else if (!strncmp(argv[idx], "-51surround=", 12))
            SOUND_51_SURROUND_set_order((u32)atoi(argv[idx] + 12));
         else if (!strcmp(argv[idx], "-51surround")) {
            fprintf(stderr, "\nERROR: -51surround option is missing the order value\n");
            return CMDLINE_ERROR;
         }
# endif
#endif
#ifdef __DVD_INPUT__
         else if (!strncmp(argv[idx], "-dvd_lang=", 10)) {
            dvd_lang[0] = *(argv[idx] + 10);
            dvd_lang[1] = *(argv[idx] + 11);
            dvd_lang[2] = '\0';
         }
         else if (!strcmp(argv[idx], "-dvd_lang")) {
            fprintf(stderr, "\nERROR: -dvd_lang option is missing a language code\n");
            return CMDLINE_ERROR;
         }
         else if (!strcmp(argv[idx], "-dvd_info"))
            params->req = DISPLAY_DVD_INFO;
         else if (!strncmp(argv[idx], "-dvd_title", 10)) {
            rc = sscanf((argv[idx] + 10), "=%hu,%hu",
                        &dvd_title, &dvd_title_last);
            if ( (!rc)||(!dvd_title) ) {
               fprintf(stderr, "\nERROR: -dvd_title option is missing the title number(s)\n");
               return CMDLINE_ERROR;
            }
            else if ( (rc == 1)||(dvd_title_last < dvd_title) )
               dvd_title_last = dvd_title;
         }
         else if (!strncmp(argv[idx], "-dvd_chapter", 12)) {
            rc = sscanf((argv[idx] + 12), "=%u", &(params->chapter));
            if ( (!rc)||(!params->chapter) ) {
               fprintf(stderr, "\nERROR: -dvd_chapter option is missing the chapter number\n");
               return CMDLINE_ERROR;
            }
         }
         else if ( (!strncmp(argv[idx], "-dvd", 4))&&
                   ( (*(argv[idx]+4) == '\0')||(*(argv[idx]+4) == '=') ) ) {
            /* open the dvd device */
            params->filename = (*(argv[idx]+4)) ? (argv[idx]+5) : dvd_device_default;
            DVD_OPEN_READ(params->dvd, params->filename);
            if (DVD_PTR_BAD(params->dvd)) {
               fprintf(stderr, "\nERROR: Cannot open dvd device %s\n", 
                       params->filename);
               params->filename = NULL;
               return CMDLINE_ERROR;
            }
            return (++idx);
         }
#endif
         else {
            fprintf(stderr, "\nERROR: Unknown option %s\n", argv[idx]);
            return CMDLINE_ERROR;
         }
      }
      else {
         /* open the input file */
         FILE_OPEN_READ(params->fp, argv[idx]);
         if (FILE_PTR_BAD(params->fp)) {
            params->filename = NULL;
            fprintf(stderr, "\nERROR: Cannot open input file %s\n", argv[idx]);
            return CMDLINE_ERROR;
         }
         params->filename = argv[idx];
         return (++idx);
      }
   } while (++idx < argc);

   fprintf(stderr, "\nERROR: No input file was given.\n");
   return CMDLINE_ERROR;
}

/* Display the usage message */
void CMDLINE_usage(char *arg0)
{
   char *name = strrchr(arg0, '/');

   if (name == NULL)
      name = arg0;
   else
      name++;
   fprintf(stderr,
           "Usage: %s [input1 options] input1 [input2 options] input2 ...\n", name);
   fprintf(stderr,
           "\nInput MPEG, or AC3 Stream from:\n"
           "  Any valid filename         Input from file or fifo\n"
#ifdef __DVD_INPUT__
           "  -dvd                       Input from default dvd device: %s\n"
           "  -dvd=[path]                Input from alternate dvd device, image, or dir.\n"
#endif
           "\nGeneral Options:\n"
           "  -help, -h                  Display this screen.\n"
           "  -loop                      Play files in a continuous loop.\n"
           "  -pids                      Display list of non-NULL PIDs in the stream.\n"
           "  -pat                       Display the Program Association Table.\n"
           "  -pmt=[n]                   Display the Program Map Table for program 'n'.\n"
           "                               Note: 'n' must be a hex value.\n"
           "  -play_prog=[n]             Play the specified program 'n'.\n"
           "                               Note: 'n' must be a hex value.\n"
           "  -play_streams=[v],[a]      Play the specified stream PIDs.\n"
           "                               v = video stream PID (use 0x1FFF to ignore)\n"
           "                               a = audio stream PID (use 0x1FFF to ignore)\n"
           "                               Note: 'v', and 'a' must be hex values.\n"
#ifdef MPEG2_SYS_CLK64           
           "  -rates                     Display list of bit rates found in the stream.\n"
#endif
           "  -start=[time]              Start playing at the given time.\n"
           "                               Note: 'time' has the format H:M:S\n"
#ifndef _NO_AUDIO_
           "\nSound Options:\n"
# ifdef __OSS_SOUND__
           "  -oss_device=[path]         Change the oss device.\n"
           "                               Current device: %s\n"
# endif
# ifdef __ALSA_SOUND__
           "  -alsa_device=[name]        Change the alsa device.\n"
           "                               Current device: %s\n"
# endif
# ifndef _NO_51_SURROUND_
           "  -51surround=[order]        Change the 5.1 surround order.\n"
           "                               'order' is one of the following values:\n"
           "                                  0 = FLFT FCEN FRGT RLFT RRGT LFE\n"
           "                                  1 = FLFT FRGT FCEN LFE  RLFT RRGT\n"
           "                                  2 = FLFT FRGT RLFT RRGT FCEN LFE\n"
           "                               Current order: %d\n"
# endif
#endif
#ifdef __DVD_INPUT__
           "\nDVD Options:\n"
           "  -dvd_chapter=[num]         Start playing at the given chapter.\n"
           "  -dvd_info                  Display dvd information.\n"
           "  -dvd_lang=[lang code]      Change language code.\n"
           "                               Current code: %s\n"
           "  -dvd_title=[num]           Play only the desired title.\n"
           "  -dvd_title=[first],[last]  Play all titles from 'first' to 'last'.\n"
#endif
           "\n"
#ifdef __DVD_INPUT__
           , dvd_device_default
#endif
#ifndef _NO_AUDIO_
# ifdef __OSS_SOUND__
           , OSS_DEVICE
# endif
# ifdef __ALSA_SOUND__
           , ALSA_DEVICE
# endif
# ifndef _NO_51_SURROUND_
           , SUR51_ORDER_DEFAULT
# endif
#endif
#ifdef __DVD_INPUT__
           , dvd_lang
#endif
           );
}

